// *****************************************************************************
//
// Copyright. 2007. Symvionics, Inc. All rights reserved.
//
// Symvionics Inc. ("Symvionics") grants a non-exclusive license for you to use
// and redistribute this software, in both source code and binary forms, and the
// accompanying documentation, if any, free of charge, provided that this
// paragraph, and the following two paragraphs appear in all copies and in all
// derivative works.  In no event shall this source code or this license thereto
// impact upon or fall within the scope of any otheragreement that you have with
// Symvionics.
//
// BY USING THIS SOFTWARE, YOU FURTHER AGREE THAT, IN NO EVENT, SHALL SYMVIONICS
// BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR
// CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS
// SOFTWARE, THE ACCOMPANYING DOCUMENTATION, IF ANY, OR ANY DERIVATIVE WORKS.
// YOU AGREE TO DEFEND, INDEMNIFY, AND HOLD HARMLESS SYMVIONICS FROM AND AGAINST
// ALL FINES, PENALTIES, LIABILITIES, DAMAGES, COSTS, AND EXPENSES INCURRED
// BY SYMVIONICS AS A RESULT OF YOUR USE OF THIS SOFTWARE INCLUDING THIRD PARTY
// USE OF ANY DERIVATIVE WORKS CREATED BY YOU.
//
// SYMVIONICS SPECIFICALLY DISCLAIMS ANY AND ALL WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE.  THE SOFTWARE AND ACCOMPANYING DOCUMENTATION LICENSED
// HEREUNDER ARE PROVIDED "AS IS" WITHOUT WARRANTY OR OBLIGATION OF SUPPORT OR
// MAINTENANCE.
//
// ***************************************************************************** 


// This file contains helper classes tAnswerCache and tIadsMultiParameterHelper. The remaining classes are support code.

#pragma warning( disable:4996 )

// Foward Decs
template <class StringType> class tString;
template <class StringType> void tOutputDebugString( StringType* debugStr, ... );
template <class ElementType, BOOL elementTypeIsAClass = FALSE> class tArray;
template <class StringType, class ConfigIfType> HRESULT tQuerySingleString( ConfigIfType* config, StringType* query, tString<StringType>* retVal );
template <class K, class D> class tHash;

// Make sure that you compile the Iads.idl and IadsTime.idl files in your project or you'll get compile errors
// Also, if you don't include the resultant Iads_i.c and IadsTime_i.c files you'll get link errors (New dev studio does this automatically?)
// If you want to simply use the "AnswerCache" class from below without including all of the IDLs into your project
// simply #define NO_AUTOMATION_FUNCTIONS before including this file
#ifndef NO_AUTOMATION_FUNCTIONS
#include "Iads.h"
#include "IadsTime.h"
#endif

#include <atlbase.h>
#include <math.h>


#ifndef _tHash_
#define _tHash_

template <class K, class D> class tHashNode
{
  public:
    K                mKey;
    D*         	     mVal;
    tHashNode<K,D>*  mNext;
    tHashNode<K,D>*  mPrev;
    tHashNode<K,D>*  mThisPointer;

    tHashNode( void ) : mNext(NULL), mPrev(NULL), mVal(NULL) { mThisPointer = this; }
    tHashNode( const K& key, D& value ): mNext(NULL), mPrev(NULL) { mThisPointer = this; mKey = key; mVal = new D(value); }
    tHashNode( const K& key, D* value ): mNext(NULL), mPrev(NULL) { mThisPointer = this; mKey = key; mVal = value; } //{ mKey = key, mVal = value; }
	  ~tHashNode()
    {
		  if ( mVal ) { delete mVal; }
		}

    void Init( const K& key, D* value )
    {
      mNext = NULL; mPrev = NULL;
      mThisPointer = this; 
      mKey = key; 
      mVal = value; 
    }

    // JFB: Careful... This is twisted... 
    void Init( const K& key, D& value )
    {
      mNext = NULL; mPrev = NULL;
      mThisPointer = this; 
      mKey = key; 
      *mVal = value; 
    }

    static BOOL isValidNode( tHashNode<K,D>* node )
    {
      try
      {
        node->mThisPointer = node->mThisPointer;
      }
      catch(...)
      {
        return FALSE;
      }
      return TRUE;
    }

};

#define tMix(a,b,c) \
{ \
  a-=b; a-=c; a^=(c>>13); \
  b-=c; b-=a; b^=(a>>8); \
  c-=a; c-=b; c^=(b>>13); \
  a-=b; a-=c; a^=(c>>12); \
  b-=c; b-=a; b^=(a>>16); \
  c-=a; c-=b; c^=(b>>5); \
  a-=b; a-=c; a^=(c>>3); \
  b-=c; b-=a; b^=(a>>10); \
  c-=a; c-=b; c^=(b>>15); \
}

#define tHashsize(n) ((unsigned int)1<<(n))
#define tHashmask(n) (tHashsize(n)-1)

template <class K, class D>
class tHash
{
  public:
    tHashNode<K,D>**    mTable;
    int                 mSize;
    int                 mCount;
    BOOL                mDeleteObjs;
    int                 mSavedHashValue;
    int                 mHashMaskValue;
    tString<char>       mTempString;
    tString<char>       mTempString2;
    K                   mResetToFirstPrimaryKey;
    BOOL                mUsedResetToFirstPrimaryKey;
    K                   mPrimaryKey; 
    
     int                 mCurrentIndex;
    tHashNode<K,D>*     mCurrentNode;

    int HashFunc( char* str )
    {
      if ( !str ) return 0;

      // Set internal state
      int a, b, c;
      register char* k = str;
      int length = tStrLen( str );
      int len = length;
      a = b = 0x9e3779b9;
      c = mSavedHashValue;

      // Process majority of key
      while ( len >= 12 )
      {
        a += (k[0]+((unsigned int)k[1]<<8)+((unsigned int)k[2]<<16)+((unsigned int)k[3]<<24));
        b += (k[4]+((unsigned int)k[5]<<8)+((unsigned int)k[6]<<16)+((unsigned int)k[7]<<24));
        c += (k[8]+((unsigned int)k[9]<<8)+((unsigned int)k[10]<<16)+((unsigned int)k[11]<<24));
        tMix( a, b, c );
        k += 12;
        len -= 12;
      }

      // Process last chars
      c += length;
      switch ( len )
      {
        case 11:
          c += ((unsigned int)k[10]<<24);
        case 10:
          c += ((unsigned int)k[9]<<16);
        case 9:
          c += ((unsigned int)k[8]<<8);
        case 8:
          b += ((unsigned int)k[7]<<24);
        case 7:
          b += ((unsigned int)k[6]<<16);
        case 6:
          b += ((unsigned int)k[5]<<8);
        case 5:
          b += k[4];
        case 4:
          a += ((unsigned int)k[3]<<24);
        case 3:
          a += ((unsigned int)k[2]<<16);
        case 2:
          a += ((unsigned int)k[1]<<8);
        case 1:
          a += k[0];
        default:
          break;
      }
      tMix( a, b, c );

      // Bitwise modulo
      c = (c & tHashmask(mHashMaskValue));
      return c;
    }

    int Hash( int in )
    {
      return in % mSize;
    }

    int Hash( unsigned int in )
    {
      return (int)(in % (unsigned int)mSize);
    }

    int Hash( __int64 in )
    {
      return (int)( (in/1000000I64/*MilliCount*/) % (__int64)mSize );
    }

    template <class StringType> int Hash( StringType* in, BOOL caseInsensitive = TRUE )
    {
      if ( caseInsensitive )
      {
        mTempString.StrCpy( in );
        mTempString.ToUpper();
        return HashFunc( mTempString.mString );  // HashFunc handles empty string
      }
      return HashFunc( in );
    }

    template <class StringType> int Hash( tString<StringType>& in, BOOL caseInsensitive = TRUE )
    {
      if ( caseInsensitive )
      {
        mTempString = in;
        mTempString.ToUpper();
        return HashFunc( mTempString.mString );  // HashFunc handles empty string
      }
      return HashFunc( in.mString );
    }

  public:
    
    tHash( BOOL deleteObjs )
    {
      mTable = NULL;
      mSize = 0;
      mCount = 0;
      mDeleteObjs = deleteObjs;
      // No real intelligence behind this value...
      mSavedHashValue = 19;
      mHashMaskValue = 0;
      mCurrentIndex = -1;
      mCurrentNode = NULL;
    }

    tHash( int sz, BOOL deleteObjs )
    {
      mTable = NULL;
      mSize = 0;
      mCount = 0;
      mDeleteObjs = deleteObjs;
      mSavedHashValue = 19;
      mHashMaskValue = 0;
      mCurrentIndex = -1;
      mCurrentNode = NULL;
      SetSize( sz );
    }

    ~tHash()
    {
      if ( mSize > 0 )
      {
        tHashNode<K,D>* tmp;
        tHashNode<K,D>* tmp2;
        for ( int ix = 0; ix < mSize; ix++ )
        {
          if ( mTable[ix] != NULL )
          {
            tmp = mTable[ix];
            while ( tmp != NULL )
            {
              tmp2 = tmp->mNext;
              if( mDeleteObjs == FALSE )
                tmp->mVal = NULL;
              
              delete tmp;
              tmp = tmp2;
            }
            mTable[ix] = NULL;
          }
        }
        delete [] mTable;
      }
    }

    HRESULT ResetToFirst( K* optionalPrimaryKey )
    {
      mCurrentNode = NULL;
      mCurrentIndex = -1;
      if ( mSize == 0 ) return S_FALSE;
      if ( optionalPrimaryKey ) 
      { 
        mResetToFirstPrimaryKey = *optionalPrimaryKey; 
        mUsedResetToFirstPrimaryKey = TRUE; 
      } 
      else 
      {
        mUsedResetToFirstPrimaryKey = FALSE;
      }

      if ( !optionalPrimaryKey )
      {
        // Need to set to the first item if no key sent in, otherwise we are trying to access an array at index -1! Not good!
        mCurrentIndex = 0; 
        register int sz = mSize;
        while ( !mCurrentNode && mCurrentIndex < sz )
        {
          if ( ( mCurrentNode = mTable[mCurrentIndex] ) != NULL )
          {
            return S_OK;
          }
          mCurrentIndex++;
        }
        mCurrentNode = NULL;
        mCurrentIndex = -1; 
        return S_FALSE;
      }

      // Jump down to the first element using the hash key
      mCurrentIndex = Hash( *optionalPrimaryKey );
      if ( mCurrentIndex < 0 ) return E_FAIL;  // Poo! Not found..

      // Now traverse to the right until we hit the first element that matches the key... 
      mCurrentNode = mTable[mCurrentIndex];   
      while ( mCurrentNode != NULL )
      {
        if ( mCurrentNode->mKey == *optionalPrimaryKey )
          return S_OK;  

        mCurrentNode = mCurrentNode->mNext;
      }

      return E_FAIL;
    }
      
    // StayWithinCurrentHashBin allows traversal only with the current bin
    D* GetNextObject( BOOL stayWithinCurrentHashBin, BOOL onlyReturnItemsThatResetToFirstPrimaryKey, K** optionalReturnKey = NULL )
    {
      if ( mSize == 0 || mCount == 0 ) return NULL;  
      if ( optionalReturnKey ) *optionalReturnKey = NULL;

      if ( mCurrentNode == NULL )
      {
        if ( stayWithinCurrentHashBin ) return NULL; // End of the line!!
        
        mCurrentIndex++;
        mCurrentNode = NULL;
        register int sz = mSize;
        while ( mCurrentIndex < sz && mCurrentNode == NULL )
        {
          mCurrentNode = mTable[mCurrentIndex];
          //#ifdef _DEBUG
          //  if ( !_CrtIsValidPointer(mCurrentNode->mVal,sizeof(mCurrentNode->mVal),TRUE) ) // Debug check for bogus entries..
          //    mCurrentIndex = mCurrentIndex;
          //#endif
          if ( mCurrentNode && !tHashNode<K,D>::isValidNode(mCurrentNode) )
          { 
            // Bad Node!!!!!
            #ifdef _ASSERTE
              _ASSERTE( FALSE );
            #endif
            mCurrentIndex++;
            continue;  
          }

          if ( mCurrentNode ) 
          {
            tHashNode<K,D>* node = mCurrentNode;
            mCurrentNode = mCurrentNode->mNext;
            // Must check for an exact match
            if ( mUsedResetToFirstPrimaryKey && !(mResetToFirstPrimaryKey == node->mKey) )
            {
              // Fast forward until the primary key is an exact match!!!
              while ( node && !(mResetToFirstPrimaryKey == node->mKey) )
              {
                node = mCurrentNode;
                if ( mCurrentNode ) mCurrentNode = mCurrentNode->mNext;
              }
              if ( !node ) 
                return NULL;
            }
            if ( optionalReturnKey ) *optionalReturnKey = &node->mKey;
            return node->mVal;
          }  
          mCurrentIndex++;
        }

        return NULL;
      }
      else
      {
        tHashNode<K,D>* node = mCurrentNode;
        mCurrentNode = mCurrentNode->mNext;  // Progress "sideways" until next is null.. then down 
        if ( mCurrentNode && !tHashNode<K,D>::isValidNode(mCurrentNode) )
        { 
          // Bad Node!!!!
          #ifdef _ASSERTE
            _ASSERTE( FALSE );
          #endif
          return NULL;  
        }
        // Must check for an exact match
        if ( mUsedResetToFirstPrimaryKey && !(mResetToFirstPrimaryKey == node->mKey) )
        {
          // Fast forward until the primary key is an exact match!!!
          while ( node && !(mResetToFirstPrimaryKey == node->mKey) )
          {
            node = mCurrentNode;
            if ( mCurrentNode ) mCurrentNode = mCurrentNode->mNext;
          }
          if ( !node ) 
            return NULL;
        }
        if ( optionalReturnKey ) *optionalReturnKey = &node->mKey;
        return node->mVal;
      }
    }

    HRESULT AddItem( K& key, D& obj, BOOL replaceWithDup, BOOL checkBothKeyAndObjForDup, int* optionalReturnBinSize )
    {
      if ( mSize == 0 )
      {
        return E_FAIL;
      }

      int index = Hash( key );
      if ( index < 0 )
      {
        return E_INVALIDARG;
      }

      tHashNode<K,D>* tmp = mTable[index];
      if ( tmp == NULL )
      {
        tHashNode<K,D>* newItem = new tHashNode<K,D>( key, obj );
        if ( !newItem ) return E_OUTOFMEMORY;
        mTable[index] = newItem;
        mCount++;
        if ( optionalReturnBinSize ) *optionalReturnBinSize = 1;
        return S_OK;
      }
      else
      {
        register int linkedListSize = 1;
        while ( tmp != NULL )
        {
          if ( tmp->mKey == key && ( !checkBothKeyAndObjForDup || *tmp->mVal == obj /*JFB:Bug!! Missing!! Won't allow matching keys*/) )
          {
            if ( replaceWithDup )
            {
              if ( mDeleteObjs )
                delete tmp->mVal;
              *tmp->mVal = obj;
              if ( optionalReturnBinSize ) *optionalReturnBinSize = linkedListSize;
              return S_OK;
            }
            else
            {
              return E_ABORT;
            }
          }

          if ( tmp->mNext == NULL )
          {
            tHashNode<K,D>* newItem = new tHashNode<K,D>( key, obj );

            // Buggy...newItem->mPrev = tmp;
            tmp->mNext = newItem;
            mCount++;
            if ( optionalReturnBinSize ) *optionalReturnBinSize = linkedListSize;
            return S_OK;
          }

          tmp = tmp->mNext;
          linkedListSize++;
        }
      }

      return S_FALSE;
    }

    HRESULT AddItem( K& key, D* obj, BOOL replaceWithDup, BOOL checkBothKeyAndObjForDup, int* optionalReturnBinSize )
    {
      if ( mSize == 0 )
      {
        return E_FAIL;
      }

      int index = Hash( key );
      if ( index < 0 )
      {
        return E_INVALIDARG;
      }

      tHashNode<K,D>* tmp = mTable[index];
      if ( tmp == NULL )
      {
        tHashNode<K,D>* newItem = new tHashNode<K,D>( key, obj );
        if ( !newItem ) return E_OUTOFMEMORY;

        mTable[index] = newItem;
        mCount++;
        if ( optionalReturnBinSize ) *optionalReturnBinSize = 1;
        return S_OK;
      }
      else
      {
        register int linkedListSize = 1;
        while ( tmp != NULL )
        {
          if ( tmp->mKey == key && ( !checkBothKeyAndObjForDup || tmp->mVal == obj /*JFB:Bug!! Missing!! Won't allow matching keys*/) )
          {
            if ( replaceWithDup )
            {
              if ( mDeleteObjs )
                delete tmp->mVal;
              tmp->mVal = obj;
              //mCount++;
              if ( optionalReturnBinSize ) *optionalReturnBinSize = linkedListSize;
              return S_OK;
            }
            else
            {
              return E_ABORT;
            }
          }

          if ( tmp->mNext == NULL )
          {
            tHashNode<K,D>* newItem = new tHashNode<K,D>( key, obj );
            if ( !newItem ) return E_OUTOFMEMORY;

            // Buggy...newItem->mPrev = tmp;
            tmp->mNext = newItem;
            mCount++;
            if ( optionalReturnBinSize ) *optionalReturnBinSize = linkedListSize;
            return S_OK;
          }

          tmp = tmp->mNext;
          linkedListSize++;
        }
      }

      return S_FALSE;
    }

    HRESULT GetItem( K& key, D*& obj )
    {
      if ( mSize == 0 )
      {
        return E_FAIL;
      }

      int index = Hash( key );
      if ( index < 0 )
      {
        return E_INVALIDARG;
      }

      tHashNode<K,D>* tmp = mTable[index];
      while ( tmp != NULL )
      {
        // NOTE: Review this for case sensitivity issues...
        if ( tmp->mKey == key )
        {
          obj = tmp->mVal;
          return S_OK;
        }

        tmp = tmp->mNext;
      }

      return S_FALSE;
    }

    D* GetItem( K& key )
    {
      if ( mSize == 0 )
      {
        return NULL;
      }

      int index = Hash( key );
      if ( index < 0 )
      {
        return NULL;
      }

      tHashNode<K,D>* tmp = mTable[index];
      while ( tmp != NULL )
      {
        // NOTE: Review this for case sensitivity issues...
        if ( tmp->mKey == key )
        {
          return tmp->mVal;
        }

        tmp = tmp->mNext;
      }

      return NULL;
    }

    HRESULT RemoveItem( K& key, BOOL deallocate )
    {
      if ( mSize == 0 )
      {
        return E_FAIL;
      }

      int index = Hash( key );
      if ( index < 0 )
      {
        tOutputDebugString( "tHash::RemoveItem, failed create Hash key to remove item\n" );
        return E_INVALIDARG;
      }

      tHashNode<K,D>* tmp = mTable[index];
      while ( tmp != NULL )
      {
        if ( tmp->mKey == key )
        {
          tHashNode<K,D>* obj = tmp;
          if ( tmp->mPrev != NULL )
          {
            tmp->mPrev->mNext = tmp->mNext;
            if ( tmp->mNext ) tmp->mNext->mPrev = tmp->mPrev;  
          }
          else
          {
            mTable[index] = tmp->mNext;   
            if ( tmp->mNext ) tmp->mNext->mPrev = mTable[index]; 
          }
          if ( !deallocate )
            tmp->mVal = NULL;

          delete tmp;
          mCount--;    
          return S_OK;
        }
        tmp = tmp->mNext;
      }

      tOutputDebugString( "tHash::RemoveItem, failed to find the item to remove\n" );
      return S_FALSE;
    }

    HRESULT RemoveItem( K& key, D* value, BOOL deallocate )
    {
      if ( mSize == 0 )
      {
        return E_FAIL;
      }

      int index = Hash( key );
      if ( index < 0 )
      {
        tOutputDebugString( "tHash::RemoveItem, failed create Hash key to remove item\n" );
        return E_INVALIDARG;
      }

      tHashNode<K,D>* tmp = mTable[index];
      tHashNode<K,D>* prev = NULL;
      while ( tmp != NULL )
      {
        if ( tmp->mKey == key && tmp->mVal == value )
        {
          //tHashNode<K,D>* obj = tmp;
          if ( !prev ) // First in list??
            mTable[index] = tmp->mNext;
          else
            prev->mNext = tmp->mNext; // Somewhere in middle or end..
          if ( !deallocate ) tmp->mVal = NULL;  

          delete tmp;
          mCount--;    
          return S_OK;
        }
        prev = tmp;
        tmp = tmp->mNext;
      }

      tOutputDebugString( "tHash::RemoveItem, failed to find the item to remove\n" );
      return S_FALSE;
    }

    K* DebugBlanketSearchForKey( D& key )
    {
      if ( mSize == 0 )
      {
        return E_FAIL;
      }

      for ( int ix = 0; ix < mSize; ix++ )
      {
        tHashNode<K,D>* tmp = mTable[ix];
        tHashNode<K,D>* prev = NULL;
        while ( tmp != NULL )
        {
          if ( tmp->mKey == value )
          {
            return tmp->mVal;
          }
          prev = tmp;
          tmp = tmp->mNext;
        }
      }
      return S_FALSE;
    }

    HRESULT DumpTableInfo()
    {
      int sz;
      int tot = 0;
      for ( int ix = 0; ix < mSize; ix++ )
      {
        sz = 0;
        tHashNode<K,D>* tmp = mTable[ix];
        while ( tmp != NULL )
        {
          sz++;
          tmp = tmp->mNext;
        }
        if ( sz != 0 )
        {
          tOutputDebugString( "Table index = %d; # nodes = %d\n", ix, sz );
          tot++;
        }
      }
      tOutputDebugString( "Total number of non-zero buckets = %d\n", tot );
      return S_OK;
    }

    int GetSize()
    {
      return mSize;
    }

    int GetCount()
    {
      return mCount;
    }

    void Clean()
    {
      if ( mSize == 0 )
      {
        return;
      }

      tHashNode<K,D>* ptr;
      tHashNode<K,D>* ptr2;
      for ( int ix = 0; ix < mSize; ix++ )
      {
        if ( (ptr = mTable[ix]) != NULL )
        {
          while ( ptr != NULL )
          {
            ptr2 = ptr;
            ptr = ptr->mNext;
            if( mDeleteObjs == FALSE )
              ptr2->mVal = NULL;

            delete ptr2;
          }
          mTable[ix] = NULL;
        }
      }
      
      delete [] mTable;
      mTable = NULL;
      mSize = 0;
      mCount = 0;

      return;
    }

    void Empty()
    {
      if ( mSize == 0 )
      {
        return;
      }

      tHashNode<K,D>* ptr;
      tHashNode<K,D>* ptr2;
      for ( int ix = 0; ix < mSize; ix++ )
      {
        if ( (ptr = mTable[ix]) != NULL )
        {
          while ( ptr != NULL )
          {
            ptr2 = ptr;
            ptr = ptr->mNext;
            if( mDeleteObjs == FALSE )
              ptr2->mVal = NULL;

            delete ptr2;
          }
          mTable[ix] = NULL;
        }
      }
      
      mCount = 0;

      return;
    }

    void SetSize( int sz )
    {
      if ( mSize > 0 )
      {
        Clean();
      }
      // Determine table size (power of 2)
      unsigned int tmpSz = sz;
      int msb = 31;
      // Find msb
      while ( msb > 0 )
      {
        if ( (tmpSz & 0x80000000)  != 0 )
        {
          break;
        }
        tmpSz = (tmpSz << 1);
        msb--;
      }

      // If sz > msb then shift
      if ( (tmpSz & 0x7fffffff) != 0 )
      {
        msb++;
      }

      // Calculate final size
      mSize = (int) pow( (double) 2.0, (double) msb );
      mHashMaskValue = msb;

      mTable = new tHashNode<K,D>*[mSize];
      memset( mTable, 0, sizeof( tHashNode<K,D>* ) * mSize );
      mCount = 0;
    }
   
};
 
#endif  //_tHash_


#include <stdio.h>  // _vsnwprintf
#ifndef _tOutputDebugString_
#define _tOutputDebugString_

template <class StringType> void tOutputDebugString( StringType* debugStr, ... )
{
  const int cMaxDebugString = 4096;

  va_list valist;
  va_start( valist, debugStr );

  if ( sizeof(StringType) == 1 )
  {
	  char DebugString[(cMaxDebugString+1)*sizeof(StringType)];  
    char* tmpBuf = const_cast<char*>(DebugString);
    _vsnprintf( tmpBuf, cMaxDebugString-1, (char*)debugStr, valist );
    tmpBuf[cMaxDebugString] = (StringType)NULL; //_vsn?printf were dropping nulls..
    OutputDebugStringA( tmpBuf );
  }
  else
  {
	  wchar_t DebugString[(cMaxDebugString+1)*sizeof(StringType)];  
    wchar_t* tmpBuf = (wchar_t*)const_cast<wchar_t*>(DebugString) ;
    _vsnwprintf( tmpBuf, cMaxDebugString-1, (wchar_t*)debugStr, valist );
    tmpBuf[cMaxDebugString] = NULL; //_vsn?printf were dropping nulls..
    OutputDebugStringW( tmpBuf );
  }

  va_end( valist );
}
#endif


////////////////////////// tArray ///////////////////////////
#ifndef _tArrayTemplates_
#define _tArrayTemplates_

#define tAbs( number ) ( (number) >= 0 ? (number) : -(number) )

template <class DestElementType, class SourceElementType, class IntType> HRESULT tCastCopy( DestElementType* d, SourceElementType* s, IntType NumberOfElements )
{
  if ( !d || !s ) return E_POINTER;
  for( register IntType i=0; i<NumberOfElements; i++ )
    d[i] = (DestElementType)s[i];
  return S_OK;
}

template <class DestElementType, class SourceElementType, class IntType> HRESULT tBitMaskCopy( DestElementType* d, SourceElementType* s, DestElementType mask, IntType NumberOfElements )
{
  if ( !d || !s ) return E_POINTER;
  register IntType nc = 0;
  for( register IntType i=0; i<NumberOfElements; i++ )
  {
    d[i] = (DestElementType)((unsigned long)s[i] & mask);
  }

  return S_OK;
}

template <class DestElementType, class SourceElementType, class IntType> HRESULT tBitSliceCopy( DestElementType* d, SourceElementType* s, IntType startBit, IntType stopBit, BOOL maintainSign, IntType NumberOfElements )
{
  if ( !d || !s ) return E_POINTER;
  register unsigned long mask = (stopBit - startBit);
  mask = (unsigned long) pow(2,mask+1) - 1;
  register IntType shift = startBit;
  mask <<= startBit;
  register IntType nc = 0;
  if ( maintainSign )
  {
    for( register IntType i=0; i<NumberOfElements; i++ )
    {
      d[i] = (DestElementType)(tAbs((long)s[i]) & mask >> shift);
      if ( s[i] < 0 )
        d[i] = -d[i];
    }
  }
  else
  {
    for( register IntType i=0; i<NumberOfElements; i++ )
    {
      d[i] = (DestElementType)((unsigned long)s[i] & mask >> shift);
    }
  }

  return S_OK;
}


template <class DestElementType, class SourceElementType, class IntType> HRESULT tClampCopy( DestElementType* d, SourceElementType* s, DestElementType destElementTypeMin, DestElementType destElementTypeMax, IntType NumberOfElements )
{
  if ( !d || !s ) return E_POINTER;
  register IntType nc = 0;
  for( register IntType i=0; i<NumberOfElements; i++ )
  {
    if ( s[i] < destElementTypeMin )
    { 
      d[i] = (DestElementType)destElementTypeMin; nc++; 
    }
    else if ( s[i] > destElementTypeMax )
    { 
      d[i] = (DestElementType)destElementTypeMax; nc++; 
    }
    else
      d[i] = (DestElementType)s[i];
  }

  return S_OK;
}

template <class ElementType, BOOL elementTypeIsAClass> 
class tArray
{
public:

  ElementType* mArray;
  long mAllocatedLen;
  ElementType mInitialValue;
  BOOL mUseInitialValueOnAllocation;
  BOOL mIsAttachedBuffer;
  long mLengthUsed;

  tArray( long initialLength = 0, BOOL zeroArray = FALSE ) 
  {
    mArray = NULL;
    mAllocatedLen = 0;
    mLengthUsed = 0;
    if ( !elementTypeIsAClass ) memset( &mInitialValue, 0, sizeof(ElementType) );
    mUseInitialValueOnAllocation = FALSE;
    mIsAttachedBuffer = FALSE;
    if ( initialLength )
    {
      mArray = new ElementType[initialLength];
      if ( mArray )
      {
        mAllocatedLen = initialLength;
        if ( zeroArray ) this->Zero();
      }
    }
  }

  ~tArray() 
  {
    if ( mArray && !mIsAttachedBuffer ) delete [] mArray;
  }

  HRESULT Attach( tArray<ElementType>& a ) 
  {
    mArray = a.mArray;
    mAllocatedLen = a.mAllocatedLen;
    mUseInitialValueOnAllocation = a.mUseInitialValueOnAllocation;
    mInitialValue = a.mInitialValue;
    mIsAttachedBuffer = TRUE;  
    return S_OK;
  }

  void Detach( void ) 
  {
    mArray = NULL;
    mAllocatedLen = 0;
    mInitialValue = 0;
    mIsAttachedBuffer = FALSE; 
  }


  HRESULT Zero( long NumberOfElements = -1 )
  {
    // Delayed allocation and initialization
    if ( !mArray )
    {
      memset( &mInitialValue, 0, sizeof(ElementType) );
      mUseInitialValueOnAllocation = TRUE;
      return S_OK;
    }      

    if ( NumberOfElements < 0 ) NumberOfElements = mAllocatedLen;
    memset( mArray, 0, sizeof(ElementType)*NumberOfElements );

    return S_OK;
  }

  HRESULT Set( ElementType value, long NumberOfElements )
  {
    // Delayed allocation and initialization
    if ( !mArray )
    {
      mInitialValue = value;
      mUseInitialValueOnAllocation = TRUE;
      return S_OK;
    }      
      
    // Need to realloc??
    if ( NumberOfElements > mAllocatedLen )
    {
      HRESULT hr = this->Realloc( NumberOfElements, FALSE );  
      if ( hr != S_OK ) return hr;
    }

    for( register long i=0; i<NumberOfElements; i++ )
      mArray[i] = value;

    return S_OK;
  }

  HRESULT MemCpy( ElementType* source, long NumberOfElements )
  {
    if ( !source ) return E_INVALIDARG;
    if ( !mArray || NumberOfElements > mAllocatedLen )
    {
      HRESULT hr = this->Realloc( NumberOfElements, FALSE );  
      if ( hr != S_OK ) return hr;
    }
    memcpy( mArray, source, sizeof(ElementType)*NumberOfElements );
    return S_OK;
  }

  HRESULT MemCpy( tArray<ElementType>& source, long NumberOfElements )
  {
    if ( !source.mArray ) return E_INVALIDARG;
    if ( !mArray || NumberOfElements > mAllocatedLen )
    {
      HRESULT hr = this->Realloc( NumberOfElements, FALSE );  
      if ( hr != S_OK ) return hr;
    }
    // Check if source large enough to hadle copy
    if ( NumberOfElements > source.mAllocatedLen )
      NumberOfElements = source.mAllocatedLen; // Clamp!!!

    memcpy( mArray, source.mArray, sizeof(ElementType)*NumberOfElements );
    return S_OK;
  }

  HRESULT Realloc( long NumberOfElements, BOOL maintainContents, long reallocPadding = 0, BOOL reallocDown = FALSE )
  {
    if ( mIsAttachedBuffer )
      return E_FAIL;  

    if ( mArray )
    {
      if ( !reallocDown )
      {
        if ( NumberOfElements <= mAllocatedLen ) return S_OK;
      }
      else
      {
        if ( NumberOfElements == mAllocatedLen ) return S_OK;
      }
    }

    NumberOfElements += reallocPadding;

    ElementType* array = new ElementType[NumberOfElements];
    if ( !array ) 
      return E_OUTOFMEMORY;

    if ( maintainContents && mArray ) 
    {
      long numToCopy = NumberOfElements < mAllocatedLen ? NumberOfElements : mAllocatedLen;
      if ( !elementTypeIsAClass )
        memcpy( array, mArray, sizeof(ElementType)*numToCopy ); 
      else
      {
        for ( register long i = 0; i < numToCopy; i++ )
          array[i] = mArray[i];   
      }

      // If initial value, set the remaining values to it...
      if ( mUseInitialValueOnAllocation )
      {
        for( register long i=numToCopy; i<NumberOfElements; i++ )
          array[i] = mInitialValue;
      }
    }
    else if ( mUseInitialValueOnAllocation )
    {
      // Delayed allocation and initialization...
      for( register long i=0; i<NumberOfElements; i++ )
        array[i] = mInitialValue;
    }

    if ( mArray ) delete [] mArray;
    mArray = array;
    mAllocatedLen = NumberOfElements;

    return S_OK;
  }

  template <class IndexType> inline ElementType& operator[](IndexType nIndex) /*const*/
  { 
    if ( nIndex < 0 )
      return mInitialValue;   

    if ( nIndex >= (IndexType)mAllocatedLen ) 
    {
      if ( mIsAttachedBuffer ) return mInitialValue;  

      register HRESULT hr = this->Realloc( nIndex+10, TRUE, 10 );
      if ( hr != S_OK ) 
        return mInitialValue;  
    }

    return mArray[nIndex]; 
  }

  inline operator ElementType*() const
	{
    return mArray;
	}

  ElementType GetElement( long index )
  {
    if ( index >= mAllocatedLen ) return (ElementType)0;
    return mArray[index];
  }

  ElementType* GetElementPtr( long index )
  {
    if ( index >= mAllocatedLen ) return NULL;
    return &mArray[index];
  }

  HRESULT SetElement( long index, ElementType e )
  {
    HRESULT hr = S_OK;
    if ( index >= mAllocatedLen )
      hr = this->Realloc( index+1, TRUE, 10 );
    if ( hr == S_OK ) mArray[index] = e;
    return hr;
  }

  HRESULT Copy( ElementType* source, long NumberOfElements )
  {
    if ( !source ) return E_INVALIDARG;
    if ( !mArray || NumberOfElements > mAllocatedLen )
    {
      HRESULT hr = this->Realloc( NumberOfElements, FALSE );  
      if ( hr != S_OK ) return hr;
    }
    memcpy( mArray, source, sizeof(ElementType)*NumberOfElements );
    return S_OK;
  }

  template <class SourceType> HRESULT CastCopy( SourceType* source, long NumberOfElements )
  {
    HRESULT hr = this->Realloc( NumberOfElements, FALSE );  if ( hr != S_OK ) return hr;
    return tCastCopy( mArray, source, NumberOfElements );
  }

  template <class SourceType> HRESULT ClampCopy( SourceType* source, long NumberOfElements, ElementType min, ElementType max )
  {
    HRESULT hr = this->Realloc( NumberOfElements, FALSE );  if ( hr != S_OK ) return hr;
    return tClampCopy( mArray, source, min, max, NumberOfElements );
  }

  HRESULT BitMaskCopy( double* source, long NumberOfElements, ElementType bitMask )
  {
    HRESULT hr = this->Realloc( NumberOfElements, FALSE );  if ( hr != S_OK ) return hr;
    return tBitMaskCopy( mArray, source, bitMask, NumberOfElements );
  }

  HRESULT BitSliceCopy( double* source, long NumberOfElements, long startBit, long stopBit, BOOL maintainSign )
  {
    HRESULT hr = this->Realloc( NumberOfElements, FALSE );  if ( hr != S_OK ) return hr;
    return tBitSliceCopy( mArray, source, startBit, stopBit, maintainSign, NumberOfElements );
  }
};

#endif   // _tArrayTemplates_



/////////////////////////////////////////// tString /////////////////////////////////////

#ifndef _tStringTemplates_
#define _tStringTemplates_

#define NO_SHLWAPI_STRFCNS // Conflict with #defines in <shlwapi.h>

// Just in case included before IadsTemplateLib.
#undef StrCpy                 
#undef StrChr                 
#undef StrRChr                
#undef StrChrI                
#undef StrRChrI               
#undef StrCmpN                
#undef StrCmpNI               
#undef StrStr                 
#undef StrStrI                
#undef StrDup                 
#undef StrRStrI               
#undef StrCSpn                
#undef StrCSpnI               
#undef StrSpn                 
#undef StrToInt               
#undef StrPBrk                
#undef StrToIntEx             
#undef StrFromTimeInterval    
#undef StrIntlEqN             
#undef StrIntlEqNI            
#undef StrFormatByteSize      
#undef StrNCat                
#undef StrTrim                
#undef ChrCmpI                


template <class OutStrType, class InStrType> HRESULT tStrConvertDup( /*[out]*/OutStrType** OutStr, /*[in]*/const InStrType* InStr )
{
  if ( OutStr == NULL || InStr == NULL ) return E_POINTER;

  if ( sizeof(OutStrType) == 1 && sizeof(InStrType) == 2 )
  {
    size_t len = wcslen((OLECHAR*)InStr) + 4;
    *OutStr = new OutStrType[ len ];
    if ( *OutStr == NULL ) return E_OUTOFMEMORY;
    wcstombs( (char*)*OutStr, (wchar_t*)InStr, len );  // Can fail?
    return S_OK;
  }
  
  if ( sizeof(OutStrType) == 2 && sizeof(InStrType) == 1 )
  {
    size_t len = strlen((char*)InStr) + 4;
    *OutStr = (OutStrType*)::SysAllocStringLen( NULL, len );
    if ( *OutStr == NULL ) return E_OUTOFMEMORY;
    mbstowcs( (wchar_t*)*OutStr, (char*)InStr, len );  // Can fail?
    return S_OK;
  }

  if ( sizeof(OutStrType) == 1 && sizeof(InStrType) == 1 )
  {
    *OutStr = (OutStrType*)(new char[ strlen((char*)InStr) + 4 ]);
    if ( *OutStr == NULL ) return E_OUTOFMEMORY;
    strcpy( (char*)*OutStr, (char*)InStr );  // Can fail?
    return S_OK;
  }

  if ( sizeof(OutStrType) == 2 && sizeof(InStrType) == 2 )
  {
    *OutStr = (OutStrType*)::SysAllocString( (OLECHAR*)InStr );
    return *OutStr == NULL ? E_OUTOFMEMORY : S_OK;
  }
}

template <class T> T* tStrCaseless( const T* wcs1, const T* wcs2 )
{
  return tStrStri( wcs1, wcs2 );
}

template <class T> T* tStrStri( const T* wcs1, const T* wcs2 )
{
  if ( !wcs1 || !wcs2 ) return NULL;

  T *cp = (T *) wcs1; 
  T *s1, *s2;
  
  while (*cp)  
  {
    s1 = cp;
    s2 = (T *) wcs2;
    if ( sizeof(T) > 1 )
    {
      while ( *s1 && *s2 && /*!(*s1-*s2)*/towupper(*s1) == towupper(*s2) )
        s1++, s2++;
    }
    else   
    {
      while ( *s1 && *s2 && /*!(*s1-*s2)*/toupper(*s1) == toupper(*s2) )
        s1++, s2++;
    }
    
    if (!*s2)
      return(cp);
    cp++;
  }
  return NULL;
}

template <class StringType> StringType* tStrChr( const StringType* st, StringType ch )
{
  if ( !st ) return NULL;
  if ( sizeof(StringType) == 1 )
    return (StringType*)strchr( (char*)st, (char)ch );
  else if ( sizeof(StringType) == 2 )
    return (StringType*)wcschr( (wchar_t*)st, (wchar_t)ch );
}

template <class StringType> StringType* tStrRChr( const StringType* st, StringType ch )
{
  if ( !st ) return NULL;
  if ( sizeof(StringType) == 1 )
    return (StringType*)strrchr( (char*)st, (char)ch );
  else if ( sizeof(StringType) == 2 )
    return (StringType*)wcsrchr( (wchar_t*)st, (wchar_t)ch );
}

template <class StringType> StringType* tStrDup( const StringType* st, /*out*/int* allocatedLen = NULL )
{
  if ( st == NULL ) return NULL;
  if ( sizeof(StringType) == 1 )
  {
    register int len = (int)strlen( (char*)st );
    char* s = new char[ len+1 ];
    if ( s != NULL ) tStrCpy( s, st );
    if ( allocatedLen ) *allocatedLen = len + 1;
    return (StringType*)s;
  }
  BSTR rt = ::SysAllocString( (BSTR)st );
  if ( allocatedLen ) *allocatedLen = ::SysStringLen( rt );
  return (StringType*)rt;
}

template <class StringType> inline HRESULT tStrFreeAndDup( StringType** dest, const StringType* src, int* outAllocatedLen = NULL )
{
  if ( dest == NULL || src == NULL ) return E_INVALIDARG;
  if ( *dest ) tStrFree( dest );
  *dest = tStrDup( src, outAllocatedLen );
  return *dest != NULL ? S_OK : E_OUTOFMEMORY;
}
 
template <class StringType> inline HRESULT tStrRealloc( StringType** oldString, int length, BOOL maintainContents, int* destLength )
{
  if ( !oldString || !destLength ) return E_POINTER;
  register StringType* newString = NULL;
  register HRESULT hr = S_OK; 
  if ( (hr=tStrAlloc(&newString,length)) != S_OK ) return hr;
  if ( maintainContents ) tStrnCpyT( newString, *oldString, *destLength ); //Gotta have destLength!  
  if ( *oldString ) tStrFree( oldString ); 
  if ( destLength ) *destLength = length;
  *oldString = newString;

  return hr;
}

template <class StringType> inline HRESULT tStrAlloc( StringType** s, int& length, int padding = 0 )
{
  register int len = length;

  if ( s == NULL ) 
    return E_POINTER;

  if ( sizeof(StringType) == 1 )
  {
    #ifdef USE_STRING_RECYCLE
    if ( gtIsStringRecycleActive )
    {
      *s = tGetRecyledString<StringType>( len );
    }
    else
    #endif
    {
      if ( len > 0 )
      {
        *s = (StringType*)new char[len+padding];
        #ifdef USE_STRING_RECYCLE
          if ( gtStringRecycleDebug ) tOutputDebugString( "Just allocated char string len = %d\n", len );
        #endif
      }
      else
      {
        *s = NULL;
        length = 0;
        #ifdef USE_STRING_RECYCLE
          if ( gtStringRecycleDebug ) tOutputDebugString( "No allocation - char string len = %d\n", len );
        #endif
        return S_OK;
      }
    }
  }
  else
  {
    #ifdef USE_STRING_RECYCLE
      if ( gtIsStringRecycleActive )
      {
        *s = tGetRecyledString<StringType>( len );
      }
      else
    #endif
    {
	    *s = (StringType*)::SysAllocStringLen( NULL, len+1+padding );
      #ifdef USE_STRING_RECYCLE
        if ( gtStringRecycleDebug ) tOutputDebugString( "Just allocated OLECHAR string len = %d\n", len+1 );
      #endif
    }
  }

  register HRESULT hr;
  if ( *s != NULL )
  {
    hr = S_OK;
    (*s)[0] = (StringType)NULL;
    length = len;
  }
  else
  {
    hr = E_OUTOFMEMORY;
    length = 0;
  }

  return hr;
}

template <class StringType> inline HRESULT tStrFree( StringType** s, int allocatedLen = 0 )
{
  if ( s == NULL ) return E_POINTER;
  if ( *s == NULL ) return S_OK;
  if ( sizeof(StringType) == 1 )
  {
    #ifdef USE_STRING_RECYCLE
    if ( gtIsStringRecycleActive && allocatedLen >= gtMinRecycledStringSize )
    {
      // Slip Length into string before recycling
      int* len = (int*)*s;
      *len = allocatedLen;
      gtStringRecycleMutex.enter();

      if ( gtStringRecycleChar.mTotalNumberOfNodes >= gtMaxRecycledStrings ) // didn't make it..
        delete [] (char*)*s;
      else
        gtStringRecycleChar.addObject( (char*)*s ); // Fucked up..gtStringRecycleChar.insertObject( gtRecycleCompareFunctionA, (char*)*s ); // Oh well... we tried...

      gtStringRecycleMutex.exit();
    }
    else
    #endif
    {
      delete [] *s;
      #ifdef USE_STRING_RECYCLE
      if ( gtStringRecycleDebug ) tOutputDebugString( "Just deleted char string len = %d\n", allocatedLen );
      #endif
    }
  }
  else
  {
    #ifdef USE_STRING_RECYCLE
    // Set minimum allocation size to 3...
    if ( gtIsStringRecycleActive && allocatedLen >= gtMinRecycledStringSize )
    {
      // Slip Length into string before recycling
      int* len = (int*)*s;
      *len = allocatedLen;
      gtStringRecycleMutex.enter();

      if ( gtStringRecycleBstr.mTotalNumberOfNodes >= gtMaxRecycledStrings ) // didn't make it..
        ::SysFreeString( (BSTR)*s );
      else
        gtStringRecycleBstr.addObject( (BSTR)*s ); //insertObject( gtRecycleCompareFunctionW, (BSTR)*s ); // Oh well... we tried...

      gtStringRecycleMutex.exit();
    }
    else
    #endif
    {
      ::SysFreeString( (BSTR)*s );
      #ifdef USE_STRING_RECYCLE
      if ( gtStringRecycleDebug ) tOutputDebugString( "Just deleted OLECHAR string len = %d\n", allocatedLen );
      #endif
    }
  }
  *s = NULL;
  return S_OK;
}
//////////////////////////////////////////////////////////


template <class StringType> inline int tStrLen( const StringType* s )
{
  if ( s == NULL ) return 0;
  if ( sizeof(StringType) == 1 )
    return (int)strlen( (char*)s );
  return (int)wcslen( (const wchar_t*)s );
}

template <class StringType> inline HRESULT tToLower( /*[in,out]*/StringType* str )
{
  if ( str == NULL ) return E_POINTER;
  if ( sizeof(StringType) == 1 )
  {
    CharLowerA( (char*)str );
    return S_OK;
  }
	CharLowerW( (wchar_t*)str );
	return S_OK;
}

template <class StringType> inline HRESULT tToUpper( /*[in,out]*/StringType* str )
{
  if ( str == NULL ) return E_POINTER;
  if ( sizeof(StringType) == 1 )
  {
    CharUpperA( (char*)str );
    return S_OK;
  }
	CharUpperW( (wchar_t*)str );
	return S_OK;
}

template <class StringType> inline StringType* tStrTok( StringType* str, const StringType* delimeters )
{
  if ( sizeof(StringType) == 1 )
    return (StringType*)strtok( (char*)str, (char*)delimeters );
  else
    return (StringType*)wcstok( (wchar_t*)str, (wchar_t*)delimeters );
}

template <class StringType> inline double tAtoF( const StringType* str )
{
  if ( str == NULL ) return 0.0;
  StringType* ErrorPos;
  if ( sizeof(StringType) == 1 )
    return strtod( (char*)str, (char**)&ErrorPos );
  else
    return wcstod( (wchar_t*)str, (wchar_t**)&ErrorPos );
}

template <class StringType> inline long tAtoI( const StringType* str, int base = 10 )
{
  if ( str == NULL ) return 0;
  StringType* ErrorPos;
  if ( sizeof(StringType) == 1 )
    return strtol( (char*)str, (char**)&ErrorPos, base );
  else
    return wcstol( (wchar_t*)str, (wchar_t**)&ErrorPos, base );
}

template <class StringType> inline unsigned long tAtoUI( const StringType* str, int base = 10 )
{
  if ( str == NULL ) return 0;
  StringType* ErrorPos;
  if ( sizeof(StringType) == 1 )
    return strtoul( (char*)str, (char**)&ErrorPos, base );
  else
    return wcstoul( (wchar_t*)str, (wchar_t**)&ErrorPos, base );
}

template <class StringType> inline __int64 tAtoI64( const StringType* str )
{
  if ( str == NULL ) return 0;
  if ( sizeof(StringType) == 1 )
  {
    return _atoi64( (char*)str );
  }
  else
  {
    register __int64 v = 0; swscanf( (wchar_t*)str, L"%I64d", &v );
    return v;
  }
}

template <class StringType> inline unsigned __int64 tAtoUI64( const StringType* str )
{
  if ( str == NULL ) return 0;
  if ( sizeof(StringType) == 1 )
  {
    register unsigned __int64 v = 0; sscanf( (char*)str, "%I64u", &v );
    return v;
  }
  else
  {
    register unsigned __int64 v = 0; swscanf( (wchar_t*)str, L"%I64u", &v );
    return v;
  }
}


template <class StringType> inline BOOL tIsDigit( StringType chr )
{
  if ( sizeof(StringType) == 1 )
    return _ismbcdigit( (unsigned int)chr );
  else
    return iswdigit( (unsigned int)chr );
}

template <class StringType> BOOL tIsInteger( const StringType* str )
{
  BOOL valid = TRUE;
  register int len = tStrLen( str );
  if ( !str || len < 1 ) return FALSE;

  for( register int chr=0; chr < len; chr++)
  {
    if ( tIsDigit(str[chr]) )
      continue;
    switch( str[chr] )
    {
      case (StringType)'-':
        if( chr > 0 )
          valid = FALSE;
        break;
      case (StringType)'+':
        if( chr > 0 )
          valid = FALSE;
        break;
      default:
        valid = FALSE;
    }
    if ( valid == FALSE ) break;
  }
  return valid;
}

template <class StringType> BOOL tIsFloat( const StringType* str )
{
  register int len  = tStrLen( str );
  if ( !str || len < 1 ) return FALSE;

  BOOL valid        = TRUE;
  int  dotCnt       = 0;
  int  minCnt       = 0;
  int  plusCnt      = 0;
  int  eCnt         = 0;
  int  decimalFound = 0;

  for( register int chr=0; chr < len; chr++ )
  {
    if( tIsDigit(str[chr]) )
      continue;

    switch( str[chr] )
    {
      case (StringType)'-':
        if( chr == 0 )
          minCnt++;
        else if( (str[chr-1]==(StringType)'e' || str[chr-1]==(StringType)'E') &&  minCnt < 2 )
          minCnt++;
        else if( (str[chr-1]==(StringType)'g' || str[chr-1]==(StringType)'G') &&  minCnt < 2 )
          minCnt++;
        else
          valid = FALSE;
        break;

      case '+':
        if( chr == 0 )
          plusCnt++;
        else if( (str[chr-1]==(StringType)'e' || str[chr-1]==(StringType)'E') && plusCnt < 2 )
          plusCnt++;
        else if( (str[chr-1]==(StringType)'g' || str[chr-1]==(StringType)'G') && plusCnt < 2 )
          plusCnt++;
        else 
          valid = FALSE;
        break;

      case '.':

        decimalFound = TRUE;
        if( ++dotCnt > 1 || eCnt )
          valid = FALSE;
        break;

      case 'e':
      case 'E':
      case 'g':
      case 'G':

        if( chr == 0 || ++eCnt > 1 )
          valid = FALSE;
        break;

      default:
        valid = FALSE;
    }

    if ( valid == FALSE ) break;
  }

  if( decimalFound == FALSE )
    valid = FALSE;

  return valid;
}

template <class StringType> inline BOOL tIsNumeric( const StringType* str )
{
  return ( tIsInteger(str) || tIsFloat(str) );
}

template <class StringType, class InputArrayType, class OutputArrayType> HRESULT tConvertCharSeparatedStringToArrays( /*[in]*/const StringType* charSeperatedTuples, /*[in]*/StringType seperatorChar, /*[in]VARENUM vtOutType,*/ /*[out]*/InputArrayType* xArray[], /*[out]*/OutputArrayType* yArray[], /*[out]*/long* lengthOfArrays )
{
  if ( !charSeperatedTuples || !xArray || !yArray || !lengthOfArrays ) return E_POINTER;

  // strtok destroys so make copy...
  StringType* copy = tStrDup( charSeperatedTuples );
  if ( copy == NULL ) return E_OUTOFMEMORY;

  // Init
  StringType* pos; StringType* tok1; StringType* tok2; StringType ss[2]; ss[0] = seperatorChar; ss[1] = (StringType)NULL;

  // First count the number of values...faster way??
  register len = 0; pos = copy;
  while ( ((tok1=tStrTok(pos,ss)) != NULL) && ((tok2=tStrTok((StringType*)NULL,ss)) != NULL) )
  {
    len++;
    pos = NULL;
  }

  *lengthOfArrays = len;
  if ( len <= 0 ) { delete [] copy; return E_INVALIDARG; }

  *xArray = new InputArrayType[len]; if ( xArray == NULL ) return E_OUTOFMEMORY;
  *yArray = new OutputArrayType[len]; if ( yArray == NULL ) { delete [] xArray; return E_OUTOFMEMORY; }

  InputArrayType* xA = *xArray;
  OutputArrayType* yA = *yArray;

  // Now grab values... (into N arrays)
  tStrCpy( copy, charSeperatedTuples );  pos = copy;
  register int index = 0;
  while ( ((tok1=tStrTok(pos,ss)) != NULL) && ((tok2=tStrTok((StringType*)NULL,ss)) != NULL) )
  {
    xA[index] = (InputArrayType)tAtoF( tok1 );   
    yA[index] = (OutputArrayType)tAtoF( tok2 );  
    index++;
    pos = NULL;
  }

  delete [] copy;
  return S_OK;
}

template <class StringType, class InputArrayType, class OutputArrayType> HRESULT tConvertCharSeparatedStringToArraysNewest( /*[in]*/StringType* charSeperatedTuples, /*[in]*/StringType seperatorChar, /*[in]VARENUM vtOutType,*/ /*[out]*/InputArrayType* xArray[], /*[out]*/OutputArrayType* yArray[], int outputVt, /*[out]*/long* lengthOfArrays )
{
  if ( !charSeperatedTuples || !xArray || !yArray || !lengthOfArrays ) return E_POINTER;

  // strtok destroys so make copy...
  tString<StringType> ct, tok1, tok2;
  ct.Attach( charSeperatedTuples );
  {
    *lengthOfArrays = 0;
    long len = ct.GetNumberOfTokens( ',' ) / 2 + 1;
    if ( len <= 0 ) return E_INVALIDARG;

    *xArray = new InputArrayType[len]; if ( xArray == NULL ) return E_OUTOFMEMORY;
    *yArray = new OutputArrayType[len]; if ( yArray == NULL ) { delete [] xArray; return E_OUTOFMEMORY; }

    InputArrayType* xA = *xArray;
    OutputArrayType* yA = *yArray;

    // Now grab values... (into N arrays)
    register int index = 0, pos = 0;

    if ( sizeof(OutputArrayType) == sizeof(CComVariant) ) // CComVariant.. so must be a string for now
    {
      CComVariant v;
      while ( ct.StrTokGood(',',&tok1,&pos) == S_OK && ct.StrTokGood(',',&tok2,&pos) == S_OK )
      {
        xA[index] = (InputArrayType)tAtoF( tok1.mString );
        v.Clear();
        v = tok2.mString;
        v.ChangeType( outputVt );
        ((CComVariant*)yA)[index] = v;   
        (*lengthOfArrays)++;
        index++;
      }
    }
    else
    {
      while ( ct.StrTokGood(',',&tok1,&pos) == S_OK && ct.StrTokGood(',',&tok2,&pos) == S_OK )
      {
        xA[index] = (InputArrayType)tAtoF( tok1.mString );
        yA[index] = (OutputArrayType)tAtoF( tok2.mString );
        (*lengthOfArrays)++;
        index++;
      }
    }
  }
  ct.Detach();

  return S_OK;
}


template <class StringType> StringType* tFGetsRealloc( /*[in,out]*/StringType** destStr, /*[in,out]*/int* destStrAllocatedLen, /*[in]*/FILE* fp )
{
  if ( destStr == NULL || destStrAllocatedLen == NULL ) return NULL;
  if ( *destStr == NULL ) // If token wasn't allocated at all...start with some reasonable size..
  {
    *destStrAllocatedLen = 200; if ( tStrAlloc(destStr,*destStrAllocatedLen) != S_OK ) return NULL;
  }

  StringType* fgetDestPointer = *destStr;
  int fgetDestLength = *destStrAllocatedLen;
  StringType* fRet = *destStr;
  while( *destStr )
  {
    StringType* end = &( (*destStr)[(*destStrAllocatedLen)-1] );
    *end = (StringType)255; // Place pattern at end...
    if ( sizeof(StringType) == 1 )
    {
      fRet = (StringType*) fgets( (char*)fgetDestPointer, fgetDestLength, fp );
    }
    else
    {
      fRet = (StringType*) fgetws( (wchar_t*)fgetDestPointer, fgetDestLength, fp );
    }
    // If pattern changed str was filled by fgets...realloc... unless it's null
    if ( *end != (StringType)255 )
    {
      if ( *(end-1) != (StringType)'\n' ) 
      {
        fgetDestLength = *destStrAllocatedLen; 
        if ( tStrRealloc( destStr, *destStrAllocatedLen*2, TRUE, destStrAllocatedLen ) != S_OK ) return NULL;
        fgetDestPointer = (*destStr) + fgetDestLength - 1;
        fgetDestLength++; // Add one for -1 above 
        *fgetDestPointer = 1;
        *fgetDestPointer = NULL;
        continue;
      }
    }
    break;
  }
  return fRet;
}


template <class StringType> StringType* tGetsRealloc( /*[in,out]*/StringType** destStr, /*[in,out]*/int* destStrAllocatedLen )
{
  if ( destStr == NULL || destStrAllocatedLen == NULL ) return NULL;
  if ( *destStr == NULL ) // If token wasn't allocated at all...start with some reasonable size..
  {
    *destStrAllocatedLen = 200; if ( tStrAlloc(destStr,*destStrAllocatedLen) != S_OK ) return NULL;
  }

  StringType* fgetDestPointer = *destStr;
  int fgetDestLength = *destStrAllocatedLen;
  StringType* fRet = *destStr;
  int readLength = 0;
  while( *destStr )
  {
    StringType* end = &( (*destStr)[(*destStrAllocatedLen)-1] );
    *end = (StringType)255; // Place pattern at end...
    if ( sizeof(StringType) == 1 )
    {
      fRet = (StringType*) fgets( (char*)fgetDestPointer, fgetDestLength, stdin );
    }
    else
    {
      fRet = (StringType*) fgetws( (wchar_t*)fgetDestPointer, fgetDestLength, stdin );
    }
    // If pattern changed str was filled by fgets...realloc... unless it's null
    if ( *end != (StringType)255 )
    {
      if ( *(end-1) != (StringType)'\n' ) 
      {
        fgetDestLength = *destStrAllocatedLen; 
        if ( tStrRealloc( destStr, *destStrAllocatedLen*2, TRUE, destStrAllocatedLen ) != S_OK ) return NULL;
        fgetDestPointer = (*destStr) + fgetDestLength - 1;
        fgetDestLength++; // Add one for -1 above 
        *fgetDestPointer = 1;
        *fgetDestPointer = NULL;
        continue;
      }
    }
    // Replace trailing linefeed with NULL...This happens automatically with gets...
    if ( ( readLength = tStrLen( fRet ) ) > 0 )
    {
      if ( fRet[readLength-1] == (StringType)'\n' )
      {
        fRet[readLength-1] = (StringType) NULL;
      }
    }
    break;
  }
  return fRet;
}


template <class T> HRESULT tReplaceCharacter( T* inStr, T srcChar, T destChar )
{
  if ( !inStr ) return E_POINTER;

  register int chr = 0;
  while( 1 )
  {
    if( inStr[chr] == (T)NULL ) break;
    if( inStr[chr] == srcChar ) inStr[chr] = destChar;
    chr++;
  }
  return S_OK;
}


/* Match candidate string to wildcard string containing any number
   of '*' or '?' wildcard characters. The '*' matches any number of
   characters, including zero characters. The '?' matches exactly
   one character. Returns TRUE if there is a match, else FALSE.

  Examples:

  WILDCARD STRING        CANDIDATE STRING
    asdfghjkl             asdfghjkl      
    *asdfgh               XXXasdfgh      
    asdfgh*               asdfghXXX      
    *aaa*bbb*ccc*         XaaaXXbbbcccXXX 
    asd??fgh??*           asdXXfghXX
    *aa??a*bbb            XXaaXXaXXbbXaaXXabbb

*/
template <class StringType> BOOL tMatchWild( const StringType* sourceStr, const StringType* wildStr )
{
  if( !wildStr || !sourceStr ) return FALSE;

  const StringType* strPtr = sourceStr;

new_segment:                                            
                                                          
  register int star = 0;                                        
  while( *wildStr == (StringType)'*' )                            
  {                                                     
    star = 1;                                          
    wildStr++;                                         
  }                                                     
                                                          
test_match:                                             
                                                          
  for( register int n=0; wildStr[n] && (wildStr[n] != (StringType)'*'); n++ )
  {                                                     
    if( wildStr[n] != strPtr[n] )                      
    {                                                   
      if( !strPtr[n] ) return FALSE;
      if( wildStr[n] == (StringType)'?' ) continue;                 
      if( !star ) return FALSE;
                                                        
      strPtr++;                                       
                                                        
      goto test_match; 
    }                                                   
  }                                                     
                                                        
  if( wildStr[n] == (StringType)'*' )                               
  {                                                     
    strPtr += n;                                      
    wildStr += n;                                      
    goto new_segment; 
  }                                                     
                                                        
  if( !strPtr[n] ) return TRUE;
  if( n && wildStr[n-1] == '*' ) return TRUE;
  if( !star ) return FALSE;
                                                        
  strPtr++;
                                                        
  goto test_match; 
}

template <class StringType> BOOL tMatchWildCaseless( const StringType* sourceStr, const StringType* wildStr )
{
  if( !wildStr || !sourceStr ) return FALSE;

  const StringType* strPtr = sourceStr;

new_segment:                                            
                                                          
  register int star = 0;                                        
  while( *wildStr == (StringType)'*' )                            
  {                                                     
    star = 1;                                          
    wildStr++;                                         
  }                                                     
                                                          
test_match:                                             
            
  register int n;
  for( n=0; wildStr[n] && (wildStr[n] != (StringType)'*'); n++ )
  {                                                     
    if( toupper(wildStr[n]) != toupper(strPtr[n]) )                      
    {                                                   
      if( !strPtr[n] ) return FALSE;
      if( wildStr[n] == (StringType)'?' ) continue;                 
      if( !star ) return FALSE;
                                                        
      strPtr++;                                       
                                                        
      goto test_match; 
    }                                                   
  }                                                     
                                                        
  if( wildStr[n] == (StringType)'*' )                               
  {                                                     
    strPtr += n;                                      
    wildStr += n;                                      
    goto new_segment; 
  }                                                     
                                                        
  if( !strPtr[n] ) return TRUE;
  if( n && wildStr[n-1] == '*' ) return TRUE;
  if( !star ) return FALSE;
                                                        
  strPtr++;
                                                        
  goto test_match; 
}


template <class StringType> HRESULT tGoodStringToken( StringType* source, /*[in,out]*/StringType** token, /*[in,out]*/int* tokenAllocatedLen, int* currentCharacterPosition, StringType delimeter )
{
  if ( source == NULL || token == NULL ) return E_POINTER;
  if ( *token == NULL ) // If token wasn't allocated at all...start with some reasonable size..
  {
    *tokenAllocatedLen = 25; if ( tStrAlloc(token,*tokenAllocatedLen) != S_OK ) return E_OUTOFMEMORY;
  }
  else if ( *tokenAllocatedLen < 25 )
  {
    *tokenAllocatedLen = 25; if ( tStrRealloc(token,*tokenAllocatedLen,FALSE,tokenAllocatedLen) != S_OK ) return E_OUTOFMEMORY;
  }

  register int destPosition = 0;

  while( source[*currentCharacterPosition] != (StringType)NULL )
  {
    if ( source[*currentCharacterPosition] != delimeter )
    {
      (*token)[destPosition++] = source[*currentCharacterPosition];
      (*currentCharacterPosition)++;
    }

    if( source[*currentCharacterPosition] == delimeter )
    {
      (*token)[destPosition] = (StringType)NULL;
      (*currentCharacterPosition)++;
      return S_OK;
    }
    else if ( source[*currentCharacterPosition] == (StringType)NULL ) 
    {
      (*token)[destPosition] = (StringType)NULL;
      return S_OK;
    }

    if ( (*currentCharacterPosition) >= (*tokenAllocatedLen-1) )
    {
      if ( tStrRealloc(token,*tokenAllocatedLen*2,TRUE,tokenAllocatedLen) != S_OK )
      {
        (*token)[destPosition] = (StringType)NULL;
        return E_OUTOFMEMORY;
      }
    }
  }

  (*token)[destPosition] = (StringType)NULL;
  return E_FAIL;
} 

template <class StringType,class IntType> HRESULT tGetTokenNumber( StringType* sourceString, const StringType* tokenToFind, StringType delimeter, IntType* zeroBasedReturnPosition, BOOL allowSubStringMatch = FALSE )
{
  if ( sourceString == NULL || tokenToFind == NULL || zeroBasedReturnPosition == NULL ) return E_POINTER;

  *zeroBasedReturnPosition = 0;
  register IntType position = 0;
  StringType* currentToken = sourceString; 

  do 
  {
    if( (sourceString[position] == delimeter) || (sourceString[position] == NULL) )
    {
      // JFBopto
      StringType origChar = sourceString[position];
      sourceString[position] = (StringType)NULL;

      if ( allowSubStringMatch == FALSE )
      {
        if ( tStrCmpi(tokenToFind,currentToken) == 0 )
        {
          sourceString[position] = origChar;
          return S_OK;
        }
      }
      else
      {
        long len = tStrLen( tokenToFind );
        if ( tStrnCmpi(tokenToFind,currentToken,len) == 0 )
        {
          sourceString[position] = origChar;
          return S_OK;
        }
      }

      sourceString[position] = (StringType)origChar;
      currentToken = &sourceString[position+1];

      (*zeroBasedReturnPosition)++; 
    }

  } while( sourceString[position++] != (StringType)NULL );

  return E_FAIL;
}

template <class StringType> HRESULT tGetTokenWild( StringType* sourceString, const StringType* wildStr )
{
  if ( sourceString == NULL || wildStr == NULL ) return E_POINTER;

  int charSize = sizeof(StringType);

  tString<char> str( sourceString );
  char* startPtr = NULL;
  char* endPtr = NULL;
  if ( str.StrStrWild( wildStr, &startPtr, &endPtr) )
  {
    if ( charSize == 1 )
    {
      strncpy((char*)sourceString, (char*)startPtr, (size_t)(endPtr-startPtr));
      sourceString[endPtr-startPtr] = '\0';
    }
    else
    {
      wcsncpy((wchar_t*)sourceString, (wchar_t*)startPtr, (size_t)((endPtr-startPtr)/charSize));
      sourceString[(endPtr-startPtr)/charSize] = '\0';
    }

    return S_OK;
  }

  sourceString[0] = '\0';
  return E_FAIL;
}

template <class StringType, class IntType> HRESULT tGetNthToken( StringType* sourceString, IntType n, StringType delimeter, /*[in,out]*/StringType** tkn, /*[in,out]*/IntType* tokenAllocatedLen )
{
  if ( sourceString == NULL || tkn == NULL || tokenAllocatedLen == NULL ) // need ? || sourceString[0] == (StringType)NULL )
    return E_FAIL;

  if ( *tkn == NULL ) // If token wasn't allocated at all...start with some reasonable size..
  {
    register int len = 25;
    if ( tStrAlloc(tkn,len) != S_OK ) return E_OUTOFMEMORY;
    *tokenAllocatedLen = len; 
  }
  else if ( *tokenAllocatedLen < 25 )
  {
    if ( tStrRealloc(tkn,25,FALSE,tokenAllocatedLen) != S_OK ) return E_OUTOFMEMORY;
  }

  register StringType* token = *tkn;  // Cache pointer for opto

  register IntType chr = 0;
  register IntType number = 0;
  StringType* start = sourceString;

  // Advance to the Nth token...
  while( (n != number) && (*start != (StringType)NULL) )
  {
    if ( *start == delimeter )
      number++;
    start++;
  }

  if ( ((n!=number)) ) //|| (number==0)) && (*start == (StringType)NULL) )
  {
    token[0] = NULL;
    return E_FAIL;
  }

  register IntType allocatedLenMinusOne = *tokenAllocatedLen - 1;  // Opto
  while( (*start != delimeter) && (*start != (StringType)NULL) )
  {
    token[chr++] = *start++;
    if ( chr >= allocatedLenMinusOne )
    {
      if ( tStrRealloc(tkn,(*tokenAllocatedLen)*2,TRUE,tokenAllocatedLen) != S_OK )
      {
        (*tkn)[chr-1] = (StringType)NULL;
        return E_OUTOFMEMORY;
      }
      token = *tkn;   // Cache pointer for opto
      allocatedLenMinusOne = *tokenAllocatedLen - 1;  // Opto
    }
  }
  token[chr] = (StringType)NULL;
  return S_OK;    
}

template <class StringType, class IntType>  HRESULT tReplaceNthToken( /*[in]*/tString<StringType>& sourceString, /*[in]*/IntType n, StringType delimeter, /*[in]*/tString<StringType>& replacementToken )
{
  return tReplaceNthToken( sourceString, n, delimeter, replacementToken.mString );
}

template <class StringType, class IntType>  HRESULT tReplaceNthToken( /*[in]*/tString<StringType>& sourceString, /*[in]*/IntType n, StringType delimeter, /*[in]*/StringType* replacementToken )
{
  // Now go to the correct position in the current row of data
  // We need to find the pointer position of the nth delimiter
  StringType* begPtr;
  StringType* endPtr;
  StringType delimeterString[2];
  delimeterString[0] = delimeter;
  delimeterString[1] = (StringType)'\0';

  // Ensure size...
  sourceString.StrRealloc( sourceString.StrLen() + tStrLen(replacementToken) + 1, TRUE );

  begPtr = tGotoNthToken( sourceString.mString, n, delimeter );
  if ( begPtr == NULL ) return E_FAIL;

  // Now get the end of the token or a null
  endPtr = tStrStr( begPtr, delimeterString );

  if ( endPtr == NULL )	// we reached the end so take the whole value
  {
    endPtr = sourceString.mString + sourceString.StrLen();
  }

  // Grab the string at the end including the delimeter, if one exists
  tString<StringType> endString;
  endString.StrCpy( endPtr );
  
  // Concatenate the replacement token at the begining of the old token
  tStrCpy( begPtr, replacementToken );

  // Concatenate the end of the source string
  sourceString.StrCat( endString );

  return S_OK;
}

template <class StringType, class IntType> HRESULT tInsertStringPos( /*[in]*/tString<StringType>& destString, /*[in]*/StringType* strToInsert, IntType characterPositionToInsert )
{
  //Handle if destString is empty, shortcut by just copying the string to insert
  if ( destString.IsEmpty() == TRUE )
  {
    destString.StrCpy(strToInsert);
    return S_OK;
  }

  StringType* begPtr;
  int destLen = destString.StrLen();

  if ( characterPositionToInsert > destLen ) characterPositionToInsert = destLen; // ??return E_INVALIDARG; // +1???

  // Ensure size...
  destString.StrRealloc( destLen + tStrLen(strToInsert) + 1, TRUE );

  begPtr = &(destString.mString[characterPositionToInsert]);

  // Grab the string to the end 
  tString<StringType> endString;
  endString.StrCpy( begPtr );
  
  // Concatenate the inserted string at the specified position
  tStrCpy( begPtr, strToInsert );

  // Concatenate the end of the source string
  destString.StrCat( endString );

  return S_OK;
}

template <class StringType> HRESULT tInsertString( /*[in]*/tString<StringType>& destString, /*[in]*/StringType* strToInsert, /*[in]*/StringType* positionToInsert )
{
  //Handle if destString is empty, shortcut by just copying the string to insert
  if ( destString.IsEmpty() == TRUE )
  {
    destString.StrCpy(strToInsert);
    return S_OK;
  }

  StringType* begPtr;
  int destLen = destString.StrLen();
  int charPos = positionToInsert-destString.mString;

  if ( charPos > destLen )
    positionToInsert = destString.mString+destLen; // ??return E_INVALIDARG; // +1???

  // Ensure size...
  destString.StrRealloc( destLen + tStrLen(strToInsert) + 1, TRUE );
  positionToInsert = destString.mString + charPos;

  begPtr = positionToInsert;

  // Grab the string to the end 
  tString<StringType> endString;
  endString.StrCpy( begPtr );
  
  // Concatenate the inserted string at the specified position
  tStrCpy( begPtr, strToInsert );

  // Concatenate the end of the source string
  destString.StrCat( endString );

  return S_OK;
}

template <class StringTypeSource, class StringTypeDest> inline HRESULT tStrCat( StringTypeDest* str, const StringTypeSource* cat )
{
  if ( str == NULL || cat == NULL ) return E_POINTER;
  
  if ( sizeof(StringTypeSource) == 1 && sizeof(StringTypeDest) == 1 )
    strcat( (char*)str, (char*)cat );
  else if ( sizeof(StringTypeSource) == 2 && sizeof(StringTypeDest) == 2 )
    wcscat( (wchar_t*)str, (wchar_t*)cat );
  else //other two cases (1->2 and 2->1)
  {
    tStrCpy(&str[tStrLen(str)], cat);
  }
  return S_OK;
}

template <class StringType> inline HRESULT tStrnCat( StringType* str, const StringType* cat, int numChars )
{
  if ( str == NULL || cat == NULL ) return E_POINTER;
  if ( sizeof(StringType) == 1 )
    strncat( (char*)str, (char*)cat, (size_t)numChars );
  else
    wcsncat( (wchar_t*)str, (wchar_t*)cat, (size_t)numChars );
  return S_OK;
}

template <class StringType> inline int tStrCmp( const StringType* str1, const StringType* str2 )
{
  if ( str1 == NULL || str2 == NULL )
  {
    if ( str1 == NULL && str2 == NULL ) return 0;
    if ( str1 == NULL && str2 != NULL ) return -1;   
    return 1;
  }
  if ( sizeof(StringType) == 1 )
    return lstrcmpA( (char*)str1, (char*)str2 );
  else
    return wcscmp( (wchar_t*)str1, (wchar_t*)str2 );
}

// Compares the trimed, caseless version of the two strings....(old strCaseCompare)
template <class StringType> int tStrCmpToken( const StringType* str1, const StringType* str2 )
{
  if ( str1 == NULL || str2 == NULL )
  {
    if ( str1 == NULL && str2 == NULL ) return 0;
    if ( str1 == NULL && str2 != NULL ) return -1;   
    return 1;
  }

/*
  // First check if the strings don't match- which in my experience with IADS
  // is the most occuring condition. this will cause a doulbe check if not equal but I
  // belive this is a resonable optimization.
#ifdef WIN32
  int retVal = strcmpi( str1, str2 ) ;    
#else
  int retVal = strcasecmp( str1, str2 ) ;
#endif
  if( retVal == 0 )
    return( 0 ) ;
*/
  // Oh well keep going....do the alltrim thing
  register int bchr1 = 0;
  register int bchr2 = 0;  

  /////////////// String 1 - find begin and end char for this string
  if ( sizeof(StringType) == 1 )
  {
    while( isspace((char)str1[bchr1]) && str1[bchr1] != (StringType)NULL )
       bchr1++;
  }
  else
  {
    while( iswspace((wchar_t)str1[bchr1]) && str1[bchr1] != (StringType)NULL )
       bchr1++;
  }

  int slen1 = tStrLen(str1) - 1 ;
  slen1 = slen1 < 0 ? 0 : slen1 ;
  if ( sizeof(StringType) == 1 )
  {
    while( isspace((char)str1[slen1]) && slen1 >= 0 )
      slen1-- ;
  }
  else
  {
    while( iswspace((wchar_t)str1[slen1]) && slen1 >= 0 )
      slen1-- ;
  }
  int trueSlen1 = (slen1-bchr1) + 1;
  //////////////
  

  ////////////// String 2 - find begin and end char for this string
  if ( sizeof(StringType) == 1 )
  {
    while( isspace((char)str2[bchr2]) && str2[bchr2] != (StringType)NULL )
       bchr2++;
  }
  else
  {
    while( iswspace((wchar_t)str2[bchr2]) && str2[bchr2] != (StringType)NULL )
       bchr2++;
  }

  int slen2 = tStrLen(str2) - 1;
  slen2 = slen2 < 0 ? 0 : slen2 ;

  if ( sizeof(StringType) == 1 )
  {
    while( isspace((char)str2[slen2]) && slen2 >= 0 )
      slen2-- ;
  }
  else
  {
    while( iswspace((wchar_t)str2[slen2]) && slen2 >= 0 )
      slen2-- ;
  }
  int trueSlen2 = (slen2-bchr2) + 1;
  ///////////////


  // upfront checks if string lengths don't match
  if( trueSlen1 < trueSlen2 )
    return( -1 ) ;
  else if( trueSlen1 > trueSlen2 )
    return( 1 ) ;

  // now do the caseless check
  if ( sizeof(StringType) == 1 )
  {
    for( register int chr=0; chr < trueSlen1; chr++ )
    {
      char strchr1 = toupper( (char)str1[bchr1+chr] ) ;
      char strchr2 = toupper( (char)str2[bchr2+chr] ) ;
      if( strchr1 != strchr2 )
        return( strchr1-strchr2 ) ;
    }
  }
  else
  {
    for( register int chr=0; chr < trueSlen1; chr++ )
    {
      wchar_t strchr1 = towupper( (wchar_t)str1[bchr1+chr] ) ;
      wchar_t strchr2 = towupper( (wchar_t)str2[bchr2+chr] ) ;
      if( strchr1 != strchr2 )
        return( strchr1-strchr2 ) ;
    }
  }

   // finally matched
  return ( 0 ) ;
}


// Doesn't check allocated length of str1
template <class StringTypeDest, class StringTypeSource> inline HRESULT tStrCpy( StringTypeDest* dest, const StringTypeSource* src, /*in,optional*/int srcAllocatedLen=-1 )
{
  if ( dest == NULL ) return E_POINTER; 
  if ( src == NULL ) { dest[0] = (StringTypeDest)NULL; return S_OK; }

  if ( sizeof(StringTypeDest) == 1 && sizeof(StringTypeSource) == 1 )
    lstrcpyA( (char*)dest, (char*)src );  // Can fail?
  else if ( sizeof(StringTypeDest) == 1 && sizeof(StringTypeSource) == 2 )
    wcstombs( (char*)dest, (wchar_t*)src, srcAllocatedLen<0 ? tStrLen(src)+1 : srcAllocatedLen );   
  else if ( sizeof(StringTypeDest) == 2 && sizeof(StringTypeSource) == 1 )
    mbstowcs( (wchar_t*)dest, (char*)src, srcAllocatedLen<0 ? tStrLen(src)+1 : srcAllocatedLen );   
  else if ( sizeof(StringTypeDest) == 2 && sizeof(StringTypeSource) == 2 )
    wcscpy( (wchar_t*)dest, (wchar_t*)src ); 
  else 
    return E_FAIL;

  return S_OK;
}  

template <class StringTypeDest, class StringTypeSource> inline HRESULT tStrnCpyT( StringTypeDest* dest, const StringTypeSource* src, int numChars )
{
  if ( dest == NULL ) return E_POINTER; 
  if ( src == NULL ) { dest[0] = (StringTypeDest)NULL; return S_OK; }

  if ( sizeof(StringTypeDest) == 1 && sizeof(StringTypeSource) == 1 )
    strncpy( (char*)dest, (char*)src, numChars );  // Can fail?
  else if ( sizeof(StringTypeDest) == 1 && sizeof(StringTypeSource) == 2 )
    wcstombs( (char*)dest, (wchar_t*)src, numChars );   
  else if ( sizeof(StringTypeDest) == 2 && sizeof(StringTypeSource) == 1 )
    mbstowcs( (wchar_t*)dest, (char*)src, numChars );   
  else if ( sizeof(StringTypeDest) == 2 && sizeof(StringTypeSource) == 2 )
    wcsncpy( (wchar_t*)dest, (wchar_t*)src, numChars ); 
  else 
    return E_FAIL;

  dest[numChars] = (StringTypeDest)NULL;
  return S_OK;
}

template <class StringType> inline int tStrCmpi( const StringType* str1, const StringType* str2 )
{
  if ( str1 == NULL || str2 == NULL )
  {
    if ( str1 == NULL && str2 == NULL ) return 0;
    if ( str1 == NULL && str2 != NULL ) return -1;   
    return 1;
  }
  if ( sizeof(StringType) == 1 )
    return _stricmp( (char*)str1, (char*)str2 );
  else
    return _wcsicmp( (wchar_t*)str1, (wchar_t*)str2 );
}

template <class StringType> inline int tStrnCmp( const StringType* str1, const StringType* str2, int n )
{
  if ( str1 == NULL || str2 == NULL )
  {
    if ( str1 == NULL && str2 == NULL ) return 0;
    if ( str1 == NULL && str2 != NULL ) return -1;   
    return 1;
  }
  if ( sizeof(StringType) == 1 )
    return strncmp( (char*)str1, (char*)str2, n );
  else
    return wcsncmp( (wchar_t*)str1, (wchar_t*)str2, n );
}

template <class StringType> inline int tStrnCmpi( const StringType* str1, const StringType* str2, int n )
{
  if ( str1 == NULL || str2 == NULL )
  {
    if ( str1 == NULL && str2 == NULL ) return 0;
    if ( str1 == NULL && str2 != NULL ) return -1;   
    return 1;
  }
  if ( sizeof(StringType) == 1 )
    return strnicmp( (char*)str1, (char*)str2, n );
  else
    return wcsnicmp( (wchar_t*)str1, (wchar_t*)str2, n );
}

template <class StringType> inline StringType* tStrStr( const StringType* str1, const StringType* str2 )
{
  if ( str1 == NULL || str2 == NULL ) return NULL; 
  if ( sizeof(StringType) == 1 )
    return (StringType*)strstr( (char*)str1, (char*)str2 );
  else
    return (StringType*)wcsstr( (wchar_t*)str1, (wchar_t*)str2 );
}

template <class StringType> BOOL tStrStrWild( const StringType* str, const StringType* wildStr, StringType** startPtr, StringType** endPtr )
{
  int charSize = sizeof(StringType);
  if (!*str) 
  { 
    if ( *wildStr == '*' ) return tStrStrWild( str, wildStr+charSize, startPtr, endPtr ); else return !*wildStr; 
  } 
  switch (*wildStr) 
  { 
  case '\0': return !*str; 
  case '*' : if ( tStrStrWild( str, wildStr+charSize, startPtr, endPtr ) ) 
               return TRUE; 
    return tStrStrWild( str+charSize, wildStr, startPtr, endPtr ); 
  case '?' : return tStrStrWild( str+charSize, wildStr+charSize, startPtr, endPtr ); 
  default  : 
    {
      BOOL ret = *wildStr == *str && tStrStrWild( str+charSize, wildStr+charSize, startPtr, endPtr );
      if ( ret )
      {
        *startPtr = (char*)str;
        if (!*endPtr )
          *endPtr = (char*)str+charSize;
      }
      return ret;
    }
  } 
}

template <class StringType>  HRESULT tCopyStringSlice( StringType* destination, StringType* start, StringType* end )
{
  if ( (destination == NULL) || (start == NULL) || (end == NULL) ) return E_POINTER;
  register int chr = 0;
  destination[0] = (StringType)NULL;
  while( ( start != end ) && ( *start != (StringType)NULL ) )
    destination[chr++] = *start++;
  destination[chr] = (StringType)NULL;
  return S_OK;
}


template <class OutStrType, class InStrType>  HRESULT tStrDupConvert( /*[out]*/OutStrType** OutStr, /*[in]*/const InStrType* InStr, /*[out,optional]*/ int* returnAllocatedLength = NULL, /*[in,optional]*/ int InStrAllocatedLen=-1, /*[in,optional]*/int reallocPadding = 8 )
{
  if ( OutStr == NULL || InStr == NULL ) return E_POINTER;
  register int aLen = InStrAllocatedLen < 0 ? tStrLen(InStr)+1 : InStrAllocatedLen; 
  HRESULT hr = tStrAlloc( OutStr, aLen, reallocPadding );  if ( hr != S_OK ) return hr;
  if ( returnAllocatedLength ) *returnAllocatedLength = aLen;
  return tStrCpy( *OutStr, InStr, aLen );
}

template <class StringType> HRESULT tAddTokenDelimeters( /*[in]*/ StringType* source, /*[out]*/StringType* dest, /*[in]*/ StringType delimeter )
{
  if( source == NULL || dest == NULL ) return E_POINTER ;

  tString<StringType> tmpStr ;
  int len = tStrLen(source) * 2 ;  // This should be bigger than needed if every other char was a delimeter char ?
  if( tmpStr.mAllocatedLen <= len ) tmpStr.StrRealloc( len+1, FALSE ) ; 

  int j = 0 ; int i = 0 ;
  while( source[i] != '\0' )
  {
    if( source[i] == delimeter )
    {
      // Delimit the delimeter with a delimeter
      tmpStr.mString[j++] = delimeter ;
    }
    tmpStr.mString[j++] = source[i++] ;
  }
  tmpStr.mString[j] = (StringType)NULL ;
  tStrCpy( dest, tmpStr.mString ) ;

  return S_OK ;
}

template <class StringType> HRESULT tRemoveTokenDelimeters( /*[in]*/ StringType* source, /*[in]*/StringType delimeter )
{
  if( !source ) return E_POINTER;

  register int i = 0;
  register char* s = source;
  register char* d = source;
  while( *s != '\0' )
  {
    if( *s == delimeter && *(s+1) == delimeter )
    {
      s++;
      continue;
    }
    *d = *s;
    s++; d++;
  }
  *d = (StringType)NULL;

  return S_OK;
}


template <class StringType> StringType* tAllTrim( /*[in]*/StringType* source )
{
  if ( source == NULL )  return NULL;

  // Catch attempts to modify a literal
  try
  {
    register int i = -1;
    register int j = 0;
    register int pos = 0;
    
    if ( sizeof(StringType) == 1 )
      while( isspace( (unsigned char)source[++i] ) && source[i] != (StringType)NULL );
    else 
      while( iswspace( (wint_t)source[++i] ) && source[i] != (StringType)NULL );
  
    while( source[i] != (StringType)NULL ) 
    {
      source[j] = source[i];
      j++;

      if( sizeof(StringType) == 1 )
      {
        if ( !isspace((unsigned char)source[i++]) ) 
          pos = j;
      }
      else if ( !iswspace((wint_t)source[i++]) )
        pos = j;
    }

    source[pos] = (StringType)NULL;
  } 
  catch (...)
  {
  }

  return source;
}

template <class StringType> StringType* tRightTrim( /*[in]*/StringType* source )
{
  if ( !source ) return NULL ;

  int len = tStrLen(source) ;

  if ( sizeof(StringType) == 1 )
    while( isspace( (char)source[--len] ) && len >= 0 ) ;
  else 
    while( iswspace( (wint_t)source[--len] ) && len > 0 ) ;

  source[len+1] = (StringType)NULL ;

  return source ;
}


template <class StringType> StringType* tAllTrimCheckMiddleWhiteSpace( /*[in]*/StringType* source, /*[out]*/BOOL* whiteSpaceInMiddle )
{
  if ( !source || !whiteSpaceInMiddle )  return NULL;

  // Catch attempts to modify a literal
  try
  {
    *whiteSpaceInMiddle = FALSE;
    register int i = -1;
    register int j = 0;
    register int pos = 0;
    register int lastWasWs = FALSE;
    
    if ( sizeof(StringType) == 1 )
      while( isspace( (char)source[++i] ) && source[i] != (StringType)NULL );
    else 
      while( iswspace( (wint_t)source[++i] ) && source[i] != (StringType)NULL );
  
    while( source[i] != (StringType)NULL ) 
    {
      source[j] = source[i];
      j++;

      if( sizeof(StringType) == 1 )
      {
        if ( !isspace((char)source[i++]) ) 
        {
          pos = j;
          if ( lastWasWs == TRUE ) *whiteSpaceInMiddle = TRUE;
          lastWasWs = FALSE;
        }
        else 
        {
          lastWasWs = TRUE;
        }
      }
      else if ( !iswspace((wint_t)source[i++]) )
        pos = j;
    }

    source[pos] = (StringType)NULL;
  } 
  catch (...)
  {
  }

  return source;
}

template <class StringType> StringType* tAllTrimSpaceTab( StringType* source )
{
  LONG i = 1, j = 0, length = 0;
  length = tStrLen( source );
  
  for (i=0; i<length; i++)
  {
    if ( source[i] == (StringType)NULL )
      break;
    
    if ( sizeof(StringType) == 1 )
    {
        if  ( isspace((unsigned char)source[i]) )
        continue;
    }
    else if ( /*sizeof(StringType) == 2 &&*/ iswspace((wint_t)source[i]) )
      continue;
    
    source[j++] = source[i];
  }
  source[j] = (StringType)NULL;
  
  return source;
}


// Get rid of trailing zeros on floats/doubles/EFormat
template <class StringType> StringType* tTrimNonSignificantZeros( StringType* s )
{
  StringType eStr[2]; eStr[0] = (StringType)'e'; eStr[1] = (StringType)NULL;  

  StringType* eInNumber = tStrStri( s, eStr );
  if ( eInNumber )
  {
    StringType* e = eInNumber;
    // Go past the 'e' and the '+' or '-'
    e++;
    if ( *e == (StringType)'+' || *e == (StringType)'-' ) e++;
    
    StringType* exponentStart = e;

    // Go to the first non zero
    while ( *e == (StringType)'0' ) e++;

    tStrCpy( exponentStart, e );

    return s;
  }

  StringType dotStr[2]; dotStr[0] = (StringType)'.'; dotStr[1] = (StringType)NULL;  
  StringType* dotInNumber = tStrStri( s, dotStr );
  if ( dotInNumber )
  {
    register long end = tStrLen( s ) - 1;
    for ( register long i = end; i > 2; i--)  //backwards loop
    {
      if ( s[i-1] == (StringType)'.' || (s[i]!=(StringType)'0' && s[i]!=(StringType)' ') ) break;
      s[i] = (StringType)NULL;
    }
    return s;
  }

  //Not 'E' format or double/float...
  return s;
}


template <class StringType>
 StringType* tStripCharacter( StringType* source, StringType characterToStrip )
{
  if ( source == NULL )
    return NULL;

  register int  s = 0;
  register int  d = 0;

  while( source[s] != (StringType)NULL )
  {
    if( source[s] != characterToStrip )
      source[d++] = source[s];
    s++;
  }

  source[d] = (StringType)NULL;

  return source;
}

template <class StringType>
 StringType* tGotoNthToken( StringType* sourceString, int tokenNumber, StringType delimeter )
{
  StringType delimeterString[2];
  delimeterString[0] = delimeter;
  delimeterString[1] = '\0';

  StringType* begPtr = sourceString;

  for (register int t = 0; t < tokenNumber; t++)
  {
    // Advance to the next instance of the delimeter
    begPtr = tStrStri( begPtr, delimeterString );
    if ( begPtr == NULL )
      return NULL;

    // Pass the delimeter
    begPtr++;
  }

  return( begPtr );
}

template <class StringType>  HRESULT tSplitPath( /*[in]*/const StringType* path, /*[in,willmodify]*/StringType* drive, /*[in,willmodify]*/StringType* dir, /*[in,willmodify]*/StringType* fileName, /*[in,willmodify]*/StringType* ext )
{
  StringType drivel[_MAX_DRIVE];
  StringType dirl[_MAX_DIR];
  StringType fnamel[_MAX_FNAME];
  StringType extl[_MAX_EXT];

  if ( sizeof(StringType) == 1 )
    _splitpath( (const char*)path, (char*)drivel, (char*)dirl, (char*)fnamel, (char*)extl );
  else
    _wsplitpath( (const wchar_t*)path, (wchar_t*)drivel, (wchar_t*)dirl, (wchar_t*)fnamel, (wchar_t*)extl );

  tStrCpy( drive, drivel );
  tStrCpy( dir, dirl );
  tStrCpy( fileName, fnamel );
  tStrCpy( ext, extl );
  tStrCpy( drive, drivel );
  
  return S_OK;
}

template <class StringType>  int tSprintf( /*[in,out]*/StringType* destString, /*[in]*/const StringType* format, ... )
{
  if ( destString == NULL ) return E_POINTER;
  va_list marker;
  va_start( marker, format );
  if ( sizeof(StringType) == 1 )
    vsprintf( (char*)destString, (const char*)format, marker ); 
  else
    vswprintf( (wchar_t*)destString, (const wchar_t*)format, marker ); 
  va_end( marker );
}


template <class StringType>
class tString
{

public:
  StringType*  mString;
  int          mAllocatedLen;


  tString()
  {
    mString = NULL; mAllocatedLen = 0;
  }

	tString( int length )
	{
    mString = NULL;  
    mAllocatedLen = length;
    tStrAlloc( &mString, mAllocatedLen );
	}

	tString( BSTR dupThisStr )
	{
    mString = NULL; mAllocatedLen = 0;
    tStrDupConvert( &mString, dupThisStr, &mAllocatedLen );
	}
	
	tString( char* dupThisStr )
	{
    mString = NULL; mAllocatedLen = 0;
    tStrDupConvert( &mString, dupThisStr, &mAllocatedLen );
	}

  tString( tString<StringType>& src )
	{
    if ( src.mAllocatedLen <= 0 && src.mString ) src.mAllocatedLen = tStrLen(src.mString) + 1;

    mString = src.StrDup();
    mAllocatedLen = tStrLen( mString ) + 1;
	}

  ~tString()
  {
    if ( mString ) tStrFree( &mString, mAllocatedLen );
  }

	HRESULT Attach( StringType* str )
	{
    if ( mString ) tStrFree( &mString, mAllocatedLen ); 
    mAllocatedLen = 0;

    mString = str;
    mAllocatedLen = tStrLen( str );  
    return S_OK;
	}

  StringType** AttachToOut()  
	{
    if ( mString ) { tStrFree( &mString, mAllocatedLen ); mAllocatedLen = 0; }
		return &mString;
	}
  void AttachToOutInit()   // JFB: Ugly but how to clean?
  {
    if ( mString )
      mAllocatedLen = tStrLen(mString) + 1;
  }

  StringType* Detach()
	{
    StringType* s = mString;
    mString = NULL;
    mAllocatedLen = 0;
    return s;
	}

  // Get internal string pointer
	StringType* GetStr() const
	{
    static StringType* emptyString = (StringType*)"\0\0";
    return mString ? mString : emptyString;
	}

	StringType* Str() const
	{
    static StringType* emptyString = (StringType*)"\0\0";
    return mString ? mString : emptyString;
	}

  // Duplicates string and returns
	StringType* StrDup()
	{
    if ( mString == (StringType*)NULL ) return tStrDup( sizeof(StringType) == 1 ? (StringType*)"" : (StringType*)L"" );
    return tStrDup( mString );
	}

  // Duplicates string and returns as BSTR
	BSTR StrDupAsBSTR()
	{
    BSTR str = NULL;
    if ( mString == (StringType*)NULL ) return tStrDup( L"" );
    tStrDupConvert( &str, mString, NULL );
    return str;
	}

  // Duplicates string and returns as char*
	char* StrDupAsCharStr()
	{
    char* str = NULL;
    if ( mString == (StringType*)NULL ) return tStrDup( "" );
    tStrDupConvert( &str, mString, NULL );
    return str;
	}

  // Length of string up to NULL
  int StrLen() const
	{
		return ( mString == NULL ? 0 : tStrLen(mString) );
	}

  // Length of memory allocated for string
  int StrAllocatedLen() const
	{
		return mAllocatedLen;
	}

  HRESULT StrRealloc( int len, BOOL maintainContents )
  {
    if ( len <= mAllocatedLen ) return S_OK;
    //mAllocatedLen = len; //don't want to set if it fails!!
    return tStrRealloc( &mString, len, maintainContents, &mAllocatedLen );
  }

  // Copy operator
	tString<StringType>& operator=(const tString<StringType>& src) { this->StrCpy(src); return *this; }
  tString<StringType>& operator=(const char* src) { this->StrCpy(src); return *this; }
	tString<StringType>& operator=(const BSTR src) { this->StrCpy(src); return *this; }

  // Append operator
	tString<StringType>& operator+=(const tString<StringType>& src) { this->StrCat(src); return *this; }
	tString<StringType>& operator+=(const char* src)	{ this->StrCat(src); return *this; }
  tString<StringType>& operator+=(const BSTR src)	{ this->StrCat(src); return *this; }

  // Boolean operators
	BOOL operator<(const tString<StringType>& src) { return (this->StrCmpi(src)<0); }
	BOOL operator<(const char* src) { return (this->StrCmpi(src)<0); }
	BOOL operator<(const BSTR src) { return (this->StrCmpi(src)<0); }
	BOOL operator>(const tString<StringType>& src) { return (this->StrCmpi(src)>0); }
	BOOL operator>(const char* src) { return (this->StrCmpi(src)>0); }
	BOOL operator>(const BSTR src) { return (this->StrCmpi(src)>0); }
	BOOL operator==(const tString<StringType>& src) { return (this->StrCmpi(src)==0); }
	BOOL operator==(const char* src) { return (this->StrCmpi(src)==0); }
	BOOL operator==(const BSTR src) { return (this->StrCmpi(src)==0); }
	BOOL operator!=(const tString<StringType>& src) { return (this->StrCmpi(src)!=0); }
	BOOL operator!=(const char* src) { return (this->StrCmpi(src)!=0); }
	BOOL operator!=(const BSTR src) { return (this->StrCmpi(src)!=0); }

  // Internal access operators
  // NOTE: The following can only be used as RHS of operations (i.e. not a reference operator)
  StringType operator[](int nIndex) const	
  { 
    _ASSERT(mString&&nIndex>-1&&nIndex<mAllocatedLen); 
    return mString[nIndex]; 
  }

  operator tString<OLECHAR>() const
  {
    return *this;
  }

  operator BSTR() const
	{
    register int sz = sizeof(StringType);
    if ( sz == 1 )
    {
      _ASSERTE( NULL ); //return (char*)mString;  // Later convert and return stack mString
       return NULL;
    }
    else if ( sz == 2 )
    {
		  return (BSTR)mString;
    }
    _ASSERTE( NULL ); // WTF happened??
    return NULL;
	}
	operator char*() const
	{
    if ( sizeof(StringType) == 1 )
    {
      return (char*)mString;
    }
    else if ( sizeof(StringType) == 2 )
    {
		  _ASSERTE( NULL );   
      return NULL;
    }
    _ASSERTE( NULL );  
    return NULL;
	}

  // Clears contents of string by inserting NULL in first character
  HRESULT StrEmpty( BOOL deallocate = FALSE )
  {
    if ( mString == NULL ) return S_OK;
    if ( deallocate ) { tStrFree(&mString,mAllocatedLen); mAllocatedLen = 0; } //this->Detach(); //We want to free the string, not detach!
    else if ( mString ) /*((mAllocatedLen > 0 )*/ mString[0] = (StringType)NULL;  // JFB: Could mAllocatedLen == 0 ??
    return S_OK;
  }

  // Is string empty? (NULL or NULL as first character)
  BOOL IsEmpty()
  {
    if ( mString == NULL ) return TRUE;
    return ( mString[0] == (StringType)NULL );
  }
  BOOL IsNotEmpty()
  {
    if ( mString == NULL ) return FALSE;
    return ( mString[0] != (StringType)NULL );
  }

  // Copy string from tString<StringType>
  HRESULT StrCpy( const tString<OLECHAR>& src, int reallocPadding = 8 )
	{
    HRESULT hr = S_OK;
    if ( src.mAllocatedLen <= mAllocatedLen )
      tStrCpy( mString, src.mString, src.mAllocatedLen/*hint*/ );
    else   
    {
      if ( mString ) tStrFree( &mString, mAllocatedLen ); 
      register int aLen = tStrLen(src.mString) + 1 + reallocPadding; // TT#4626
      hr = tStrDupConvert( &mString, src.mString, &mAllocatedLen, aLen, reallocPadding );
    }
		return hr;
	}

  HRESULT StrCpy( const tString<char>& src, int reallocPadding = 8 )
	{
    HRESULT hr = S_OK;
    if ( src.mAllocatedLen <= mAllocatedLen )
      tStrCpy( mString, src.mString, src.mAllocatedLen/*hint*/ );
    else   
    {
      if ( mString ) tStrFree( &mString, mAllocatedLen ); 
      register int aLen = tStrLen(src.mString) + 1 + reallocPadding; // TT#4626
      hr = tStrDupConvert( &mString, src.mString, &mAllocatedLen, aLen, reallocPadding );
    }
		return hr;
	}

  // Copy string from BSTR and convert if necessary
  HRESULT StrCpy( const BSTR src, int reallocPadding = 8 )
	{
    if ( src == NULL ) return StrEmpty();
    HRESULT hr = S_OK;
    register int aLen = tStrLen(src) + 1;
    if ( aLen <= mAllocatedLen )
      hr = tStrCpy( mString, src, aLen );
    else
    {
      if ( mString ) tStrFree( &mString, mAllocatedLen ); 
      hr = tStrDupConvert( &mString, src, &mAllocatedLen, aLen, reallocPadding );
    }
		return hr;
	}

  // Copy string from char* and convert if necessary
  HRESULT StrCpy( const char* src, int reallocPadding = 8 )
	{
    if ( src == NULL ) return StrEmpty();
    HRESULT hr = S_OK;
    register int aLen = tStrLen(src) + 1;
    if ( aLen <= mAllocatedLen )
      tStrCpy( mString, src, aLen );
    else
    {
      if ( mString ) tStrFree( &mString, mAllocatedLen ); 
      hr = tStrDupConvert( &mString, src, &mAllocatedLen, aLen, reallocPadding );
    }
		return hr;
	}

  //These 2 versions will always null-terminate the dest string
  HRESULT StrnCpyT( const StringType* src, int numChars )
  {
    if ( src == NULL ) return StrEmpty();
    if ( numChars >= mAllocatedLen ) 
    { 
      HRESULT hr = tStrRealloc( &mString, numChars+1, FALSE, &mAllocatedLen ); 
      if ( hr != S_OK ) return hr; 
    }
		return tStrnCpyT( mString, src, numChars );
  }

  HRESULT StrnCpyT( const tString<StringType>& src, int numChars )
  {
    if ( numChars >= mAllocatedLen ) 
    { 
      HRESULT hr = tStrRealloc( &mString, numChars+1, FALSE, &mAllocatedLen ); 
      if ( hr != S_OK ) return hr; 
    }
		return tStrnCpyT( mString, src.mString, numChars );
  }

  // Duplicates string slice and returns
  HRESULT CopyStringSlice( /*[in]*/StringType* StartCharPos, /*[in]*/StringType* EndCharPos )
  {
    register int lengthNeeded = (int)(EndCharPos - StartCharPos + 1) + 1;
    if ( mAllocatedLen < lengthNeeded && tStrRealloc(&mString,lengthNeeded,FALSE,&mAllocatedLen) != S_OK ) return E_OUTOFMEMORY;
    return tCopyStringSlice( mString, StartCharPos, EndCharPos );
  }

  // Concatenate string, rellocate if necessary
	HRESULT StrCat( /*[in]*/const tString<StringType>& str, /*[in]*/int reallocPadding = 40 )
	{
    register int meLen = this->StrLen();
    register int strLen = str.StrLen();
    if ( (meLen+strLen) < mAllocatedLen )
		  return tStrCat( mString, str.mString );
    else
    {
      HRESULT hr = tStrRealloc( &mString, meLen+strLen+1+reallocPadding, TRUE, &mAllocatedLen ); if (hr!=S_OK) return hr;
      //mAllocatedLen = meLen + strLen + 1 + reallocPadding;
      return tStrCat( mString, str.mString );
    }
	}

  HRESULT StrCat( /*[in]*/const OLECHAR* str, /*[in]*/int reallocPadding = 40 )
  {
    register int meLen = this->StrLen();
    register int strLen = tStrLen(str);
    if ( (meLen+strLen) < mAllocatedLen )
		  return tStrCat( mString, str );
    else
    {
      HRESULT hr = tStrRealloc( &mString, meLen+strLen+1+reallocPadding, TRUE, &mAllocatedLen ); if (hr!=S_OK) return hr;
      //mAllocatedLen = meLen + strLen + 1 + reallocPadding;
      return tStrCat( mString, str );
    }
  }

  HRESULT StrCat( /*[in]*/const char* str, /*[in]*/int reallocPadding = 40 )
  {
    register int meLen = this->StrLen();
    register int strLen = tStrLen(str);
    if ( (meLen+strLen) < mAllocatedLen )
		  return tStrCat( mString, str );
    else
    {
      HRESULT hr = tStrRealloc( &mString, meLen+strLen+1+reallocPadding, TRUE, &mAllocatedLen ); if (hr!=S_OK) return hr;
      //mAllocatedLen = meLen + strLen + 1 + reallocPadding;
      return tStrCat( mString, str );
    }
  }

  // These 2 versions will always null-terminate the dest string
  HRESULT StrnCat( const StringType* src, int numChars )
  {
    if ( src == NULL ) return S_OK;
    if ( ( numChars + this->StrLen() ) >= mAllocatedLen ) 
    { 
      HRESULT hr = tStrRealloc( &mString, numChars+this->StrLen()+1, FALSE, &mAllocatedLen ); 
      if ( hr != S_OK ) return hr; 
    }
		return tStrnCat( mString, src, numChars );
  }

  HRESULT StrnCat( const tString<StringType>& src, int numChars )
  {
    if ( src.mString == NULL ) return S_OK;
    if ( ( numChars + this->StrLen() ) >= mAllocatedLen ) 
    { 
      HRESULT hr = tStrRealloc( &mString, numChars+this->StrLen()+1, FALSE, &mAllocatedLen ); 
      if ( hr != S_OK ) return hr; 
    }
		return tStrnCat( mString, src.mString, numChars );
  }

  // Compare string and return -1,0,1
  int StrCmp( const StringType* str )
  {
    return tStrCmp( mString, str );
  }

  // Compare n characters of string and return -1,0,1
  int StrnCmp( const StringType* str, int n )
  {
    return tStrnCmp( mString, str, n );
  }

  // Compare string case insensative and return -1,0,1
  int StrCmpi( const StringType* str )
  {
    return tStrCmpi( mString, str );
  }
  int StrCmpi( const tString<StringType>& str )
  {
    return tStrCmpi( mString, str.mString );
  }

  // Compare n characters of string case insensative and return -1,0,1
  int StrnCmpi( const StringType* str, int n )
  {
    return tStrnCmpi( mString, str, n );
  }
  
  // Find substring and return pointer to position 
  StringType* StrStr( const StringType* str )
  {
    return tStrStr( mString, str );
  }

  // Find wildcarded substring and modify startPtr and endPtr to reflect the position of the string
  BOOL StrStrWild( const StringType* wildStr, StringType** startPtr, StringType** endPtr )
  {
    return tStrStrWild( mString, wildStr, startPtr, endPtr );
  }

  StringType* StrChr( StringType ch )
  {
    if ( !mString ) return NULL;
    if ( sizeof(StringType) == 1 )
      return (StringType*)strchr( (char*)mString, (char)ch );
    else if ( sizeof(StringType) == 2 )
      return (StringType*)wcschr( (wchar_t*)mString, (wchar_t)ch );
  }

  StringType* StrRChr( StringType ch )
  {
    if ( !mString ) return NULL;
    if ( sizeof(StringType) == 1 )
      return (StringType*)strrchr( (char*)mString, (char)ch );
    else if ( sizeof(StringType) == 2 )
      return (StringType*)wcsrchr( (wchar_t*)mString, (wchar_t)ch );
  }

  // Find substring case insensative and return pointer to position 
  StringType* StrStrCaseless( const StringType* str )
  {
    return tStrStri( mString, str );
  }

  StringType* StrStri( const StringType* str )
  {
    return tStrStri( mString, str );
  }

  // Replace every instance of character in string with another
  HRESULT ReplaceCharacter( StringType replaceThisChar, StringType withThisChar )
  {
    return tReplaceCharacter( mString, replaceThisChar, withThisChar );
  }

  // Replace one/every instance of a string with another, case sensative
  HRESULT ReplaceString( StringType* replaceThisString, StringType* withThisString, BOOL everyInstance )
  {
    StringType* start;
    HRESULT hr = E_FAIL;
    register StringType* pos = mString;
    tString<StringType> endPart;
    register int lenWithThis = tStrLen( withThisString );
    register int lenThis = tStrLen( replaceThisString );
    register int nextPosOffset = 0;
    while ( start = tStrStr(pos,replaceThisString) )
    {
      hr = S_OK;
      nextPosOffset += lenWithThis + (int)(start-pos);
      endPart.StrCpy( start+lenThis );
      *start = (StringType)NULL;
      this->StrCat( withThisString );
      this->StrCat( endPart ); 
      // JFB: Carefull here StrCats could realloc... don't use pointers...
      pos = this->mString + nextPosOffset;
      if ( !everyInstance ) break;
    }
    return hr;
  }

  // Replace one/every instance of a string with another, case insensative
  HRESULT ReplaceStringI( StringType* replaceThisString, StringType* withThisString, BOOL everyInstance )
  {
    StringType* start;
    HRESULT hr = E_FAIL;
    register StringType* pos = mString;
    tString<StringType> endPart;
    register int lenWithThis = tStrLen( withThisString );
    register int lenThis = tStrLen( replaceThisString );
    register int nextPosOffset = 0;
    while ( start = tStrStri(pos,replaceThisString) )
    {
      hr = S_OK;
      nextPosOffset += lenWithThis + (int)(start-pos);
      endPart.StrCpy( start+lenThis );
      *start = (StringType)NULL;
      this->StrCat( withThisString );
      this->StrCat( endPart ); 
      // JFB: Carefull here StrCats could realloc... don't use pointers...
      pos = this->mString + nextPosOffset;
      if ( !everyInstance ) break;
    }
    return hr;
  }


  // Match candidate string to wildcard string containing any number of '*' or '?' wildcard characters
  BOOL MatchWild( const StringType* wildStr )
  {
    return tMatchWild( mString, wildStr );
  }

  HRESULT MakeValidParameterName(void)
  {
    if ( mString == NULL ) return E_FAIL ;

    static char  mIllegalSet[] = {' ', '/', '\\', ':','+', '-', '*', '\'', '^', '&', '<', '>', '?', '|', '"', ',', '\0' } ;

    int len = tStrLen( mString ) ;
    for( register int n=0; n < len; n++ )
    {
      for( register int cs=0; mIllegalSet[cs] != '\0'; cs++ )
      {
        if( mString[n] == (StringType)mIllegalSet[cs] )
          mString[n] = '_' ;
      }
    }
    return S_OK ;
  }

  HRESULT MakeValidFileName(void)
  {
    if ( mString == NULL ) return E_FAIL ;

    static char mIllegalFileSet[] = {'"', '\\', '/', ':', '*', '?', '<', '>', '|', '\0' } ;

    int len = tStrLen( mString ) ;
    for( register int n=0; n < len; n++ )
    {
      for( register int cs=0; mIllegalFileSet[cs] != '\0'; cs++ )
      {
        if( mString[n] == (StringType)mIllegalFileSet[cs] )
          mString[n] = '_' ;
      }
    }
    return S_OK ;
  }

  HRESULT RemoveCharacters( int StartCharPos, int NumChars )
  {
    if ( mString == NULL ) return E_FAIL;
    if ( StartCharPos >= mAllocatedLen || StartCharPos+NumChars >= mAllocatedLen )
      return E_INVALIDARG; 
    if ( StartCharPos == 0 )
      return tStrCpy( mString, &mString[NumChars] ) != NULL ? S_OK : E_FAIL;
    return tStrCpy( &mString[StartCharPos], &mString[StartCharPos+NumChars] );
    //return E_FAIL;  // JFB: Later... must also ensure NumChars less than strlen..See 2nd line above...
  }

  HRESULT RemoveCharactersFromEnd( int numChars )
  {
    register int pos = this->StrLen() - numChars;
    if ( pos < 0 || !mString ) return E_FAIL;
    mString[pos] = (StringType)NULL;
    return S_OK;
  }

  HRESULT RemoveCharactersFromLeft( StringType atThisDelimeter, BOOL removeDelimeter = TRUE )
  {
    StringType* d = this->StrChr( atThisDelimeter );
    if ( !d ) return E_FAIL;
    return this->StrCpy( removeDelimeter?d+1:d );
  }
  HRESULT RemoveCharactersFromRight( StringType pastThisDelimeter, BOOL removeDelimeter = TRUE )
  {
    StringType* d = this->StrChr( pastThisDelimeter );
    if ( !d ) return E_FAIL;
    d = removeDelimeter ? d : d+1;
    *d = (StringType)NULL;
    return S_OK;
  }

  HRESULT InsertString( StringType* stringToInsert, int characterPositionToInsert )
  {
    return tInsertStringPos( *this, stringToInsert, characterPositionToInsert );
  }
  HRESULT InsertString( StringType* stringToInsert, StringType* postionToInsert )
  {
    return tInsertString( *this, stringToInsert, postionToInsert );
  }

  // Compare the ws-trimmed, case insensative versions of this & str2
  inline int StrCmpToken( /*[in]*/const StringType* str2 )
  {
    return tStrCmpToken( mString, str2 );
  }

  // Tokenize string given delimeter. Set currentCharacterPosition to zero at start
  HRESULT StrTokGood( /*[in]*/StringType delimeter, /*[in,out]*/tString<StringType>* token, /*[in,out]*/int* currentCharacterPosition )
  {
    return tGoodStringToken( mString, &(token->mString), &(token->mAllocatedLen), currentCharacterPosition, delimeter );
  }

  // Find and return a token position (0..NumberOfTokens) within string of tokens
  HRESULT GetTokenNumber( /*[in]*/const StringType* tokenToFind, /*[in]*/StringType delimeter, /*[out]*/int* tokenNum, BOOL allowSubStringMatch = FALSE )
  {
    return tGetTokenNumber( mString, tokenToFind, delimeter, tokenNum, allowSubStringMatch );
  }

  HRESULT GetTokenWild( /*[in]*/const StringType* wildStr )
  {
    return tGetTokenWild( mString, wildStr );
  }

  // Find and return a token by position (0..NumberOfTokens) within string of tokens
  HRESULT GetNthToken( /*[in]*/int tokenNumber, /*[in]*/StringType delimeter, /*[in,willModify]*/tString<StringType>* token )
  {
    return tGetNthToken( mString, tokenNumber, delimeter, &(token->mString), &(token->mAllocatedLen) );
  }
  HRESULT GetNthToken( /*[in]*/int tokenNumber, /*[in]*/StringType delimeter, /*[in,willModify]*/tString<StringType>& token )
  {
    return tGetNthToken( mString, tokenNumber, delimeter, &(token.mString), &(token.mAllocatedLen) );
  }

  HRESULT ReplaceNthToken( /*[in]*/int tokenNumber, /*[in]*/StringType delimeter, /*[in]*/tString<StringType>& replacementToken )
  {
    return tReplaceNthToken( *this, tokenNumber, delimeter, replacementToken );
  }

  HRESULT ReplaceNthToken( /*[in]*/int tokenNumber, /*[in]*/StringType delimeter, /*[in]*/StringType* replacementToken )
  {
    return tReplaceNthToken( *this, tokenNumber, delimeter, replacementToken );
  }

  // Find and return a pointer to a token within string of tokens
  StringType* GotoNthToken( /*[in]*/int tokenNumber, /*[in]*/StringType delimeter )
  {
    return tGotoNthToken( mString, tokenNumber, delimeter );
  }

  // Remove any leading or trailing white-space
  StringType* AllTrim()
	{
    return tAllTrim( mString );
  }

  // Add double delimeter characters has protection from parser code
  HRESULT AddTokenDelimeters(StringType delimeter, tString<StringType>* dest)
	{
    if( dest ) dest->StrRealloc( mAllocatedLen*2, FALSE ) ;
    
    return tAddTokenDelimeters( mString, dest?dest->mString:NULL, delimeter ) ;
  }

  // Remove double delimeter characters has protection from parser code
  HRESULT RemoveTokenDelimeters( StringType delimeter )
	{
    return tRemoveTokenDelimeters( mString, delimeter );
  }

  // Remove any leading or trailing white-space
  StringType* RightTrim()
	{
    return tRightTrim( mString );
  }

  StringType* TrimNonSignificantZeros()
  {
    return tTrimNonSignificantZeros( mString );
  }

  StringType* AllTrimSpaceTab()
	{
    return tAllTrimSpaceTab( mString );
  }


  // Remove any instance of character within the string
  StringType* StripCharacter( StringType characterToStrip )
	{
    return tStripCharacter( mString, characterToStrip );
  }

  // Convert all characters in string to lower case
  HRESULT ToLower()
	{
    if ( mString == NULL ) return S_OK;
    return tToLower( mString );
  }

  // Convert all characters in string to upper case
  HRESULT ToUpper()
  {
    if ( mString == NULL ) return S_OK;
    return tToUpper( mString );
  }

  // Convert string to double
  double AtoF()
  {
    return tAtoF( mString );
  }

  // Convert string to long
  long AtoI()
  {
    return tAtoI( mString );
  }

  // Convert string to unsigned long
  unsigned long AtoUI()
  {
    return tAtoUI( mString );
  }

  __int64 AtoI64()
  {
    return tAtoI64( mString );
  }

  unsigned __int64 AtoUI64()
  {
    return tAtoUI64( mString );
  }

  // Is string a numerical representation
  BOOL IsNumeric()
  {
    return tIsNumeric( mString );
  }

  // Is string a floating point representation
  BOOL IsFloat()
  {
    return tIsFloat( mString );
  }

  // Is string a integer representation
  BOOL IsInteger()
  {
    return tIsInteger( mString );
  }

  // Set string formatted output... Careful not to use format arguments of other than StringType! (Don't mix BSTRs/chars)
  HRESULT Sprintf( /*[in]*/const StringType* format, ... )
  {
    if ( mString == NULL || mAllocatedLen < 1 )
    {
      mAllocatedLen = 40; // Some size to start with...
      if ( tStrAlloc(&mString,mAllocatedLen) != S_OK ) 
      { 
        mAllocatedLen = 0; return E_OUTOFMEMORY; 
      } 
    }

    va_list marker;
    va_start( marker, format );
    register long sz = 0;
    // JFB: Use tSprintf template above later...
    if ( sizeof(StringType) == 1 )
    {
      while ( (sz=_vsnprintf((char*)mString,mAllocatedLen-1,(const char*)format,marker)) == -1 )
      {
        if ( tStrRealloc( &mString, mAllocatedLen*2, FALSE, &mAllocatedLen ) != S_OK ) return E_OUTOFMEMORY;
        //mAllocatedLen *= 2;
      }
    }
    else
    {
      while ( (sz=_vsnwprintf((wchar_t*)mString,mAllocatedLen-1,(const wchar_t*)format,marker)) == -1 )
      {
        if ( tStrRealloc( &mString, mAllocatedLen*2, FALSE, &mAllocatedLen ) != S_OK ) return E_OUTOFMEMORY;
        //mAllocatedLen *= 2;
      }
    }
    if ( mString ) mString[sz] = (StringType)NULL; //_vsn?printf were dropping nulls..
    va_end( marker );

    return S_OK;
  }

  // Set string to binary representation of data
  HRESULT SprintfBinary( unsigned long data, int numberOfBits )
  {
    if ( numberOfBits < 0 ) numberOfBits = 0;

    register int strLenNeeded = numberOfBits + 1;
    if ( mAllocatedLen < strLenNeeded )
    {
      tStrRealloc( &mString, strLenNeeded, FALSE, &mAllocatedLen );
      //mAllocatedLen = strLenNeeded;
      if ( mString == NULL ) { mAllocatedLen = 0; return E_OUTOFMEMORY; }
    }
    mString[numberOfBits] = (StringType)NULL;
    for( register int i=(numberOfBits-1); i >= 0; i-- )
    {
      if( data & 1 )
        mString[i] = (StringType)'1';
      else
        mString[i] = (StringType)'0';
      data >>= 1;
    }
    return S_OK;
  }

  // Pass in null for parts not wanted..
  HRESULT SplitPath( /*[in,willModify]*/tString<StringType>* drive, /*[in,willModify]*/tString<StringType>* dir, /*[in,willModify]*/tString<StringType>* fileName, /*[in,willModify]*/tString<StringType>* ext )
  {
    // drive = _MAX_DRIVE   dir = _MAX_DIR   fileName = _MAX_FNAME   ext = _MAX_EXT;
    if ( drive ) drive->StrRealloc( _MAX_DRIVE*2, FALSE );
    if ( dir ) dir->StrRealloc( _MAX_DIR*10, FALSE );
    if ( fileName ) fileName->StrRealloc( _MAX_FNAME*2, FALSE );
    if ( ext ) ext->StrRealloc( _MAX_EXT*2, FALSE );

    return tSplitPath( mString, drive?drive->mString:NULL, dir?dir->mString:NULL, fileName?fileName->mString:NULL, ext?ext->mString:NULL );
  }

  HRESULT SplitFileName( /*[in,willModify]*/tString<StringType>* driveAndDir, /*[in,willModify]*/tString<StringType>* fileNameAndExt )
  {
    if ( !mString ) return E_FAIL;

    StringType* fn = tStrRChr( mString, (StringType)'\\' );
    if ( fn ) 
    {
      if ( fileNameAndExt ) { fileNameAndExt->StrCpy( fn+1 ); fileNameAndExt->AllTrim(); }
      if ( driveAndDir ) driveAndDir->CopyStringSlice( mString, fn+1/*Include BackHack*/ );
      return S_OK;
    }
    else 
    {
      if ( fileNameAndExt ) { fileNameAndExt->StrCpy( mString ); fileNameAndExt->AllTrim(); }
      if ( driveAndDir ) driveAndDir->StrEmpty();
      return S_OK;
    }
  }

  HRESULT MakeTempFileName( /*[in]*/StringType* fileName )
  {
    this->StrRealloc( (_MAX_DRIVE+_MAX_DIR+_MAX_FNAME+_MAX_EXT+1)*4, FALSE ); 

    if ( sizeof(StringType) == 1 )
      ::GetTempPath( (_MAX_DRIVE+_MAX_DIR+_MAX_FNAME+_MAX_EXT+1)*4, (char*)this->mString );
    else
      ::GetTempPathW( (_MAX_DRIVE+_MAX_DIR+_MAX_FNAME+_MAX_EXT+1)*4, (BSTR)this->mString );

    StringType* fn = tStrRChr( fileName, (StringType)'\\' );
    this->StrCat( fn ? fn+1 : fileName ); 
    this->AllTrim();

    return S_OK; 
  }

  HRESULT ConvertToArrays( /*[in]*/StringType seperatorChar, /*[out]*/double* xArray[], /*[out]*/double* yArray[], /*[out]*/long* lengthOfArrays )
  {
    return tConvertCharSeparatedStringToArrays( mString, seperatorChar, xArray, yArray, lengthOfArrays );
  }

  int GetNumberOfMatchingSubStrings( const StringType* SubString )
  {
    if ( mString == NULL ) return 0;
	  register int num = 0;
    StringType* inThisString = mString;
	  while( (inThisString=tStrStr(inThisString,SubString)) != NULL )
	  {
		  inThisString++;
		  num++;
	  }
	  return num;
  }

  int GetNumberOfTokens( StringType delimeter )
  {
    if ( !mString ) return 0;
    register int nt = 0;
    register StringType* p = mString;
    while ( *p )
    {
      if ( *p++ == delimeter ) nt++;
    }
    return nt+1;//don't forget the plus 1 ( 1 pipe = 2 tokens, etc. )
  }

  tString<StringType> CreateExtendedFileName( const StringType* directory, const StringType* fileName )
  {
    //safety in case directory and filename are NULL...Needed when method returns (*this)!
    this->StrCpy("");

    this->StrCpy( directory );
    if ( directory != NULL && directory[0] != (StringType)NULL )
    {
      StringType lastChar = directory[tStrLen(directory)-1];
      if ( lastChar != (StringType)'\\' )
      this->StrCat( "\\" );
    }
    this->StrCat( fileName );
    
    return (*this);
  }

  StringType* FGets( FILE* fp )
  {
    return tFGetsRealloc( &mString, &mAllocatedLen, fp );
  }

  StringType* Gets()
  {
    return tGetsRealloc( &mString, &mAllocatedLen );
  }

  HRESULT SetIrigTime( __int64 irigTime, int digitsPastDecimal = 3, BOOL showDays = TRUE )
  {
    const __int64 tSecondToTimeConversionFactor = (__int64) 1000000000;
    const __int64 tHourToTimeConversionFactor = (__int64) 3600000000000;
    const __int64 tMaxInternalMinorTime = (__int64) 86400000000000;
    const __int64 tMinuteToTimeConversionFactor = (__int64) 60000000000;

    int days = (int)(irigTime / tMaxInternalMinorTime);
    __int64 minorTime = irigTime % tMaxInternalMinorTime;
    int hours = (int)(minorTime /  tHourToTimeConversionFactor) % 24;   
    int minutes = (int)(minorTime / tMinuteToTimeConversionFactor) % 60;
    int seconds =  (int)(minorTime / tSecondToTimeConversionFactor) % 60;
    const __int64 tNanosecondsInSecond = (__int64) 1000000000;
    digitsPastDecimal = tClamp( digitsPastDecimal, 0, 9 );
    int partialSeconds = (int)( minorTime - (minorTime / tNanosecondsInSecond * tNanosecondsInSecond) );
    partialSeconds = (int)( (double)partialSeconds * pow(10.0,digitsPastDecimal-9) );

    if ( sizeof(StringType) == 1 )
    {
      char formatStr[50];
      if ( showDays )
      {
        sprintf( formatStr, "%%03d:%%02d:%%02d:%%02d.%%0%dd", digitsPastDecimal );
        if ( digitsPastDecimal != 0 )
          return this->Sprintf( (StringType*)formatStr, days, hours, minutes, seconds, partialSeconds );
        else
          return this->Sprintf( (StringType*)"%03d:%02d:%02d:%02d", days, hours, minutes, seconds );
      }
      else
      {
        sprintf( formatStr, "%%02d:%%02d:%%02d.%%0%dd", digitsPastDecimal );
        if ( digitsPastDecimal != 0 )
          return this->Sprintf( (StringType*)formatStr, hours, minutes, seconds, partialSeconds );
        else
          return this->Sprintf( (StringType*)"%02d:%02d:%02d", hours, minutes, seconds );
      }
    }
    else
    {
      OLECHAR formatStr[50];
      if ( showDays )
      {
        swprintf( formatStr, L"%%03d:%%02d:%%02d:%%02d.%%0%dd", digitsPastDecimal );
        if ( digitsPastDecimal != 0 )
          return this->Sprintf( (StringType*)formatStr, days, hours, minutes, seconds, partialSeconds );
        else
          return this->Sprintf( (StringType*)"%03d:%02d:%02d:%02d", days, hours, minutes, seconds );
      }
      else
      {
        swprintf( formatStr, L"%%02d:%%02d:%%02d.%%0%dd", digitsPastDecimal );
        if ( digitsPastDecimal != 0 )
          return this->Sprintf( (StringType*)formatStr, hours, minutes, seconds, partialSeconds );
        else
          return this->Sprintf( (StringType*)"%02d:%02d:%02d", hours, minutes, seconds );
      }
    }
  }

  __int64 GetIrigTime()
  {
    const __int64 tSecondToTimeConversionFactor = (__int64) 1000000000;
    const __int64 tHourToTimeConversionFactor = (__int64) 3600000000000;
    const __int64 tMaxInternalMinorTime = (__int64) 86400000000000;
    const __int64 tMinuteToTimeConversionFactor = (__int64) 60000000000;
    const __int64 tMillisecondToTimeConversionFactor = (__int64) 1000000;

    int s = 0; int m = 0; int h = 0; int d = 0;
    StringType* period = this->StrStr( "." );
    tString<StringType> majorTimeString;
    tString<StringType> minorTimeString;
    if( period != NULL )
    {
      majorTimeString.CopyStringSlice( mString, period );
      period++;
      minorTimeString.StrCpy( period );
    }
    else
    {
      majorTimeString.StrCpy( mString );
      minorTimeString.StrCpy( "0" );
    }

    sscanf( majorTimeString.mString, "%d:%d:%d:%d", &d, &h, &m, &s );
  
    __int64 timeInIadsCounts = ( (__int64)s + ((__int64)60*(__int64)m) + ((__int64)60*(__int64)60*(__int64)h) ) * (__int64)1E9;
    int lenMinor = minorTimeString.StrLen();
    if ( lenMinor > 9 ) minorTimeString.mString[9] = '\0';
    timeInIadsCounts += minorTimeString.AtoI() * (__int64)pow( 10.0, 9-lenMinor ); 
    timeInIadsCounts += d * 24 * tHourToTimeConversionFactor;

    return timeInIadsCounts;
  }

  HRESULT DisectSqlSelectQuery( /*[out,optional]*/tString<StringType>* selectTokens, /*[out,optional]*/tString<StringType>* fromToken, /*[out,optional]*/tString<StringType>* whereClause )
  {
    if ( !mString ) return E_UNEXPECTED;
    if ( sizeof(StringType) > 1 ) return E_NOTIMPL; // Not yet... How to templatize string constants?

    StringType* selectPtr = tStrStri( mString, "select " );
    if( !selectPtr ) return E_FAIL;

    StringType* fromPtr = tStrStri( mString, "from " );
    if( !fromPtr ) return E_FAIL;

    // No where statement is OK! Everything is true!
    StringType* wherePtr = tStrStri( mString, "where " );

    // Copy all of the characters between the end of the select clause 
    // and the begining of the where clause. This is the 
    if ( selectTokens ) 
    {
      selectTokens->CopyStringSlice( selectPtr+tStrLen("select"), fromPtr );
    }
 
    // Get the From tokens and the where equation
    if( !wherePtr )
    {
      if ( fromToken ) fromToken->StrCpy( fromPtr+tStrLen("from") );
    
      // DEBUG: No where clause is OK...
      if ( whereClause ) whereClause->StrCpy( "" );
    }
    else 
    {
      if ( fromToken ) fromToken->CopyStringSlice( fromPtr+tStrLen("from"), wherePtr );
      if ( whereClause ) whereClause->StrCpy( wherePtr+tStrLen("where") ) ;
    }

    if ( selectTokens ) selectTokens->AllTrim();
    if ( whereClause ) whereClause->AllTrim();
    if ( fromToken ) fromToken->AllTrim();
    
    return S_OK;
  }

  HRESULT DisectSqlUpdateQuery( /*[out,optional]*/tString<StringType>* table, /*[out,optional]*/tString<StringType>* setTokens, /*[out,optional]*/tString<StringType>* whereClause )
  {
    if ( !mString ) return E_UNEXPECTED;
    if ( sizeof(StringType) > 1 ) return E_NOTIMPL; // Not yet... How to templatize string constants?

    StringType* updatePtr = tStrStri( mString, "update " ) ;
    if( updatePtr == NULL )
    {
      if ( ( updatePtr = tStrStri( mString, "updat2 " ) ) == NULL )
      {
        if ( ( updatePtr = tStrStri( mString, "updat3 " ) ) == NULL )
        {
          tOutputDebugString( "tString: Illegal update statement, missing 'update' clause (%s)\n", mString );
          return E_FAIL;
        }
      }
    }

    char* setPtr = tStrStri( mString, "set " ) ;
    if( setPtr == NULL )
    {
      tOutputDebugString( "tString: Illegal update statement, missing 'set' clause (%s)\n", mString );
      return E_FAIL ;
    }

    char* whereStart = mString;
    char* wherePtr = NULL;

    while( *whereStart != '\0' )
    {
      char* nextWhere = tStrStri( whereStart, "where " );
      if ( nextWhere == NULL ) break;
      wherePtr = nextWhere;
      whereStart = nextWhere + 6; //strlen("where ");
    }

    // Copy the update table name 
    if ( table )
    {
      table->CopyStringSlice( updatePtr+7, setPtr-1 ); 
      table->AllTrim();
    }

    // Grab the where equation and the set tokens... No where clause is OK!
    if( wherePtr == NULL )
    {
      if ( whereClause ) whereClause->StrCpy( "0" ); 
      if ( setTokens ) setTokens->StrCpy( setPtr+4 );  
    }
    else 
    {
      // Copy the set tokens.. DEBUG: for now only "*"
      if ( setTokens )    
      {
        setTokens->CopyStringSlice( setPtr+4, wherePtr-1 ); 
        setTokens->AllTrim();
      }
      if ( whereClause ) 
      {
        whereClause->StrCpy( wherePtr+6 ); 
        whereClause->AllTrim();
      }
    }

    return S_OK;
  }

};

#endif // StringTemplates


#ifndef NO_AUTOMATION_FUNCTIONS


#ifndef _tIadsMultiParameterHelper_
#define _tIadsMultiParameterHelper_

template <class StringType> class tIadsMultiParameterHelper
{
public:

  bool mWasInitialized;
  tString<StringType> mDataGroup;
  tArray< IParameter* > mInputParameters;
  CComPtr<IIadsTime> mIadsTime;
  double mIterationSampleRate;

  tIadsMultiParameterHelper( void )
  {
    ::CoInitialize( NULL );
    mInputParameters.Zero();
    mWasInitialized = false;
    mIterationSampleRate = -1.0;
    ::CoCreateInstance( CLSID_IadsTime, NULL, CLSCTX_INPROC_SERVER, IID_IIadsTime, (void**)&mIadsTime );
  }

  ~tIadsMultiParameterHelper( void )
  {
    ::CoUninitialize();
    for( register int i=0; i<mInputParameters.mLengthUsed; i++ )
    {
      if ( mInputParameters[i] ) mInputParameters[i]->Release();
    }
  }

  HRESULT InitializeWithDataGroup( StringType* dataGroup, BOOL ErrOnMissingParm = TRUE )
  {
    for( register int i=0; i<mInputParameters.mLengthUsed; i++ )
    {
      if ( mInputParameters[i] ) mInputParameters[i]->Release();
    }
    mInputParameters.mLengthUsed = 0;

    // Get a handle to the IADS application
    CComPtr<IApplication> iads;
    CLSID clsid = CLSID_NULL;
    ::CLSIDFromProgID( L"Iads.Application", &clsid );
    HRESULT hr = ::CoCreateInstance( clsid, NULL, CLSCTX_INPROC_SERVER, IID_IApplication, (void**)&iads );
    if ( hr != S_OK || !iads ) return E_FAIL;

    // Grab the configuration
    CComPtr<IConfiguration> iadsConfig;
    hr = iads->get_Configuration( &iadsConfig );
    if ( hr != S_OK ) return hr;

    // Query for the data group name given
    tString<OLECHAR> dg = dataGroup; dg.AllTrim();
    tString<OLECHAR> q; q.Sprintf( L"select Parameters from DataGroups where Name = '%s'", dg.GetStr() );

    tString<OLECHAR> parameterList;
    hr = tQuerySingleString<OLECHAR,IConfiguration>( iadsConfig.p, q.GetStr(), &parameterList );
    if ( hr != S_OK ) return hr;

    tString<OLECHAR> paramName; int pos = 0;  
    while ( parameterList.StrTokGood(L',',&paramName,&pos) == S_OK )
    {
      HRESULT hr = iadsConfig->CreateParameter( paramName, &mInputParameters[mInputParameters.mLengthUsed] );
      if ( hr != S_OK )
      {
        mInputParameters[mInputParameters.mLengthUsed] = NULL;
        tOutputDebugString( L"tIadsMultiParameterHelper: Unable to create parameter (%s) in dataGroup (%s)\n", paramName.GetStr(), dg.GetStr() );
        if( ErrOnMissingParm )
          return E_INVALIDARG; // JFB/JW: Quit here if user specifies
      }
      mInputParameters.mLengthUsed++;
    }

    mDataGroup.StrCpy( dataGroup );
    mWasInitialized = true;
    return S_OK;
  }

  HRESULT InitializeWithSprintfFormat( StringType* DataGroup )
  {
    // JFB: Later..
    return E_NOTIMPL;
  }

  HRESULT SetIterationSampleRate( double sampleRate = -1.0/*DefaultSetToHighestInGroup*/ )
  {
    if ( sampleRate < 0.0 )
    {
      for( register int i=0; i<mInputParameters.mLengthUsed; i++ )
      {
        if ( !mInputParameters[i] ) continue;
        register double sr;
        HRESULT hr = mInputParameters[i]->get_SampleRate( &sr );
        if ( hr != S_OK ) continue;
        if ( sr > sampleRate ) sampleRate = sr; 
      }
    }
    if ( sampleRate < 0.0 ) return E_FAIL;
    for( register int i=0; i<mInputParameters.mLengthUsed; i++ )
    {
      if ( !mInputParameters[i] ) continue;
      HRESULT hr = mInputParameters[i]->put_InterpolatedSampleRate( sampleRate );
      if ( hr != S_OK ) return hr;
    }
    mIterationSampleRate = sampleRate;
    return S_OK;
  }

  HRESULT SetCurrentTime( double currentTimeOfYear )
  {
    if ( !mIadsTime ) return E_FAIL;

    register __int64 currentTimeNanos = (__int64)( currentTimeOfYear * 1E9 );
    mIadsTime->put_TotalNanoseconds( currentTimeNanos );

    return this->SetCurrentTime( mIadsTime );
  }

  HRESULT SetCurrentTime( IIadsTime* currentTimeOfYear )
  {
    if ( !mIadsTime ) return E_FAIL;
    if ( currentTimeOfYear != mIadsTime )
      mIadsTime->SetTime( currentTimeOfYear );

    register HRESULT hr, returnHr = S_OK;
    register IParameter* currentParameter;
    for( register int n=0; n < mInputParameters.mLengthUsed; n++ )
    {
      currentParameter = mInputParameters[n];

      // JFB: Shouldn't happen!
      if ( !currentParameter ) continue;

      hr = currentParameter->put_CurrentTime( mIadsTime );
      if ( hr != S_OK )
      {
        returnHr = hr;
      }
    }

    return returnHr;
  }

  double GetCurrentValue( int index )
  {
    register IParameter* currentParameter = mInputParameters[index];
    if ( !currentParameter ) return 0.0;

    register double value;
    register HRESULT hr = currentParameter->Value( iadsTimeNoChange, &value );
    if ( hr != S_OK )
      return 0.0;
    return value;
  }

};

#endif  // _tIadsMultiParameterHelper_


#ifndef _ConfigTemplates_
#define _ConfigTemplates_

template <class StringType, class ConfigIfType> HRESULT tQuerySingleString( ConfigIfType* config, StringType* query, tString<StringType>* retVal )
{
  if ( !config || !query || !retVal ) return E_POINTER;

  CComPtr<IQueryResults> qrs;
  CComPtr<IQueryResult> qr;

  tString<OLECHAR> q( query );
  HRESULT hr = config->SqlQuery( q, &qrs );
  if ( hr != S_OK || !qrs.p ) return hr;

  VARIANT v; v.vt = VT_I2; v.iVal = 0;
  CComVariant vResult;  

  qrs->Item( v, &qr );
  if ( !qr.p ) return S_FALSE;

  hr = qr->Item( v, &vResult );  
  if ( hr != S_OK || vResult.vt == VT_EMPTY ) return S_FALSE;
  if ( vResult.vt != VT_BSTR ) 
  {
    hr = vResult.ChangeType( VT_BSTR );
    if ( hr != S_OK ) return hr;
  }

  retVal->StrCpy( vResult.bstrVal );
  return S_OK;
}

template <class StringType> class tIadsConfigHelper
{
public:

  CComPtr<IConfiguration> mIadsConfig;
  tString<StringType> mReturnString;

	tIadsConfigHelper()
	{
	  // Get a handle to the IADS application
	  CComPtr<IApplication> iads;
	  CLSID clsid = CLSID_NULL;
	  ::CLSIDFromProgID( L"Iads.Application", &clsid );
	  HRESULT hr = ::CoCreateInstance( clsid, NULL, CLSCTX_INPROC_SERVER, IID_IApplication, (void**)&iads );
	  if ( hr != S_OK || !iads ) return;

	  // Grab the configuration
	  hr = iads->get_Configuration( &mIadsConfig );
	  if ( hr != S_OK ) return;
	}

	double GetNumericConstant( StringType* constName )
	{
    if ( !mIadsConfig.p ) return 0.0;

	  // Query for the data group name given
	  tString<OLECHAR> cn = constName; cn.AllTrim();
	  tString<OLECHAR> q; q.Sprintf( L"select Value from Constants where ConstantName = '%s'", cn.GetStr() );

	  tString<OLECHAR> constValue;
	  HRESULT hr = tQuerySingleString<OLECHAR,IConfiguration>( mIadsConfig.p, q.GetStr(), &constValue );
	  if ( hr != S_OK ) return 0.0;

	  return constValue.AtoF();
	}

	StringType* GetStringConstant( StringType* constName )
	{
    mReturnString.StrEmpty();
    if ( !mIadsConfig.p ) return mReturnString.GetStr();

	  // Query for the data group name given
	  tString<OLECHAR> cn = constName; cn.AllTrim();
	  tString<OLECHAR> q; q.Sprintf( L"select Value from Constants where ConstantName = '%s'", cn.GetStr() );

	  tString<OLECHAR> constValue;
	  HRESULT hr = tQuerySingleString<OLECHAR,IConfiguration>( mIadsConfig.p, q.GetStr(), &constValue );
	  if ( hr != S_OK ) return mReturnString.GetStr();

    mReturnString.StrCpy( constValue );
	  return mReturnString.GetStr();
	}

};

#endif // _ConfigTemplates_

#endif // #ifndef NO_AUTOMATION_FUNCTIONS


#ifndef _tAnswerCache_
#define _tAnswerCache_

// Used to cache answers for COM functions performing costly computations as well as MultiInMultiOut
template <class OutputDataType> class tAnswerCache
{
public:
	int                mHashRingIndex;
  bool               mRingIsFull;
  tHash < __int64, OutputDataType > mAnswerCache;
  __int64*           mTimeHashRing;
  OutputDataType*    mAnswers;  // Ring the hash table... Used so we don't need to reallocate hash elements
  int                mAnswerArraySize;

  tAnswerCache( long numberOfAnswersToCache ) : mAnswerCache( FALSE/*DeleteObjects*/ ) // Call with NUM_ANSWERS_TO_CACHE
  {
    mTimeHashRing = NULL;
    mAnswers = NULL;
    SetCacheSize( numberOfAnswersToCache );
  }

  // Returns NULL if no answers cached at given time
  OutputDataType* GetAnswers( __int64 atThisTime )
  {
    // Search the hash table to see if hash key currentTimeNanos has an object for that time.
    OutputDataType* answers;
    if ( mAnswerCache.GetItem(atThisTime,answers) == S_OK )
    {
      return answers;
    }
    return NULL;
  }

  // Call this to get a free answer buffer from the cache
  OutputDataType* GetNextFreeAnswerBuffer( void )
  {
	  // Get ready to save the new entry into the cache/ring
    if( ++mHashRingIndex >= mAnswerArraySize )
    {
      mRingIsFull = true;
      mHashRingIndex = 0;
    }

    // Get the next free answer array/struct from ring buffer pool
    return &mAnswers[mHashRingIndex]; 
  }

  // Call this to add answers to the cache
  HRESULT AddAnswers( __int64 atThisTime, OutputDataType* answerToAdd )
  {
    // Check if the Hash ring is full... If so, we need to start blowing entries out of the hash (oldest ones first)
    if ( mRingIsFull )
    {
      // Remove the oldest entry from the hash cache ;)
      mAnswerCache.RemoveItem( mTimeHashRing[mHashRingIndex], &mAnswers[mHashRingIndex], FALSE );
    }

 	  // Now add the current answers to the cache
    mTimeHashRing[mHashRingIndex] = atThisTime;
 	  return mAnswerCache.AddItem( atThisTime, answerToAdd, TRUE, FALSE, NULL );
  }

  // Call this to change the cache size dynamically
  HRESULT SetCacheSize( long numberOfAnswersToCache )
  {
		mHashRingIndex = -1;
		mRingIsFull = false;
    if ( mTimeHashRing ) delete mTimeHashRing;
    mTimeHashRing = new __int64[ numberOfAnswersToCache+5 ];
    if ( mAnswers ) delete mAnswers;
    mAnswers = new OutputDataType[ numberOfAnswersToCache+5 ];
 	  mAnswerCache.SetSize( numberOfAnswersToCache );
    mAnswerArraySize = numberOfAnswersToCache;
    return S_OK;
  }

};
#endif // _tAnswerCache_





