// SampleDataGroupFunction.h : Declaration of the CSampleDataGroupFunction

#pragma once
#include "resource.h"       // main symbols

#include "SampleFunctionVS2005.h"


#if defined(_WIN32_WCE) && !defined(_CE_DCOM) && !defined(_CE_ALLOW_SINGLE_THREADED_OBJECTS_IN_MTA)
#error "Single-threaded COM objects are not properly supported on Windows CE platform, such as the Windows Mobile platforms that do not include full DCOM support. Define _CE_ALLOW_SINGLE_THREADED_OBJECTS_IN_MTA to force ATL to support creating single-thread COM object's and allow use of it's single-threaded COM object implementations. The threading model in your rgs file was set to 'Free' as that is the only threading model supported in non DCOM Windows CE platforms."
#endif

#include "IadsFunction.h"
#include "IadsComFunctionHelper.h"  // Don't forget to include the helper template for this function

// CSampleDataGroupFunction

class ATL_NO_VTABLE CSampleDataGroupFunction :
	public CComObjectRootEx<CComSingleThreadModel>,
	public CComCoClass<CSampleDataGroupFunction, &CLSID_SampleDataGroupFunction>,
	public IDispatchImpl<ISampleDataGroupFunction, &IID_ISampleDataGroupFunction, &LIBID_IadsFunctionLib, /*wMajor =*/ 1, /*wMinor =*/ 0>,
  public IDispatchImpl<IIadsFunction, &__uuidof(IIadsFunction), &LIBID_IadsFunctionLib, /* wMajor = */ 1, /* wMinor = */ 0>
{
public:
	CSampleDataGroupFunction()
	{
	}

DECLARE_REGISTRY_RESOURCEID(IDR_SAMPLEDATAGROUPFUNCTION)


BEGIN_COM_MAP(CSampleDataGroupFunction)
	COM_INTERFACE_ENTRY(ISampleDataGroupFunction)
	COM_INTERFACE_ENTRY(IDispatch)
	COM_INTERFACE_ENTRY(IIadsFunction)
END_COM_MAP()



	DECLARE_PROTECT_FINAL_CONSTRUCT()

	HRESULT FinalConstruct()
	{
    // This variable is for performance and initialization reasons as you will see below in the Compute function. 
    // Make sure to add this member variable to your class -> bool mWasInitialized
    mWasInitialized = false;

		return S_OK;
	}

	void FinalRelease()
	{
	}

  STDMETHOD(Compute)( VARIANT * dataIn,  VARIANT * dataOut)
  {
    // About this function
    // FunctionName in Iads -> SampleFunctionVC2005.SampleDataGroupFunction  (Basically, this is the "ProgId" for this class)
    // How to call in Iads  -> SampleFunctionVC2005.SampleDataGroupFunction( "DataGroup", CurrentTimeOfYear() )   (You can also feed in constant values) 
    // In Args              -> All parameters defined in the datagroup (first argument). See Iads help file for a discussion of data groups
    // Out Args             -> FloatValueSum

    // Background info on the Compute function interface
    // The VARIANT "dataIn" argument contains a "SafeArray" of input arguments passed from the user's
    // function (from of their derived equation) into this function. It will contain all the values
    // that you'll need to compute the function result. The result is returned in the "dataOut" argument.
    // The "dataOut" arguement is a VARIANT so we can return numeric results, as well as strings and other types
    // For a more advanced example on returning strings and other types, see the SimpleFunction2 class in this project
    // This function will be called once per data point calculated

    // Get the input arguments array
    VARIANT* argsArray = (VARIANT*)(dataIn->parray->pvData);  // Could use SafeArrayAccessData, but slower..

    // The very first call to this function is designated as an "initializion call". 
    // Check all parameter types here and then never do again. Adding this check to your code will speed up the performance 
    if ( !mWasInitialized )
    {
      // First Step: We need to check how many arguments were passed into our function from the user
      // Example, if the user creates a derived parameter with the function -> SampleFunctionVC2005.SampleDataGroupFunction( "DataGroup", CurrentTimeOfYear() )
      // The argument count would be 2
      int argCount = dataIn->parray->rgsabound->cElements;
      if ( argCount != 2 )
      {
        return DISP_E_BADPARAMCOUNT;
      }

      // Second Step: Check Types of each arg..... Either VT_R8 (floating point value), VT_BSTR (string value) for now...
      // In IADS, most every type of numerical argument is passed via an 8 byte floating point value VT_R8. Strings are passed as type VT_BSTR.
      // If you're in doubt, use VT_R8. You can also break here in the code and examine the argsArray[N].vt value to see the actual type 
      // Example, if the user creates a derived parameter with the function -> SampleFunctionVC2005.SampleDataGroupFunction( "DataGroup", CurrentTimeOfYear() )
      // The argument type for the first argument would be VT_BSTR and the second VT_R8    
      if ( argsArray[0].vt != VT_BSTR ) return E_INVALIDARG;
      if ( argsArray[1].vt != VT_R8 ) return E_INVALIDARG;

      // Third step: Get the actual first arg by extracting from the array of input argument and use this arg (string DataGroup)
      // to initialize the multi-parameter helper
      register BSTR dataGroupName = argsArray[0].bstrVal;
      HRESULT hr = mMultiParameterHelper.InitializeWithDataGroup( dataGroupName, TRUE/*ReturnErrorIfParamsMissing*/ );

      // If the paramter helper fails initialization then it's most likely unable to find one or more parameters in the datagroup
      // Return an error here so the user can fix the problem.
      if ( hr != S_OK ) 
        return E_INVALIDARG;

      mWasInitialized = true;
      return S_OK;
    }

    // Fourth step: Get the actual values of each arg by extracting from the array of input arguments
    register BSTR dataGroupName = argsArray[0].bstrVal;
    register double currentTimeOfYear = argsArray[1].dblVal;

    // Fifth step: Get the actual values of each arg by extracting from the array of input arguments
    mMultiParameterHelper.SetCurrentTime( currentTimeOfYear );

    // Final step: Perform your function's purpose and return the output value
    // Add em up and ship it out... Because we're returning a number, the return type is VT_R8 (double) for now... Iads will convert if necessary..
    double sum = 0.0;
    for( register int i=0; i<mMultiParameterHelper.mInputParameters.mLengthUsed; i++ )
    {
      sum += mMultiParameterHelper.GetCurrentValue( i );
    }
    dataOut->vt = VT_R8;
    dataOut->dblVal = sum;

    return S_OK;
  }

private:

  tIadsMultiParameterHelper<OLECHAR> mMultiParameterHelper;
  bool mWasInitialized;

};

OBJECT_ENTRY_AUTO(__uuidof(SampleDataGroupFunction), CSampleDataGroupFunction)
