// SimpleFunction.h : Declaration of the CSimpleFunction

#pragma once
#include "resource.h"       // main symbols

#include "SampleFunctionVS2005.h"
#include "IadsFunction.h"


#if defined(_WIN32_WCE) && !defined(_CE_DCOM) && !defined(_CE_ALLOW_SINGLE_THREADED_OBJECTS_IN_MTA)
#error "Single-threaded COM objects are not properly supported on Windows CE platform, such as the Windows Mobile platforms that do not include full DCOM support. Define _CE_ALLOW_SINGLE_THREADED_OBJECTS_IN_MTA to force ATL to support creating single-thread COM object's and allow use of it's single-threaded COM object implementations. The threading model in your rgs file was set to 'Free' as that is the only threading model supported in non DCOM Windows CE platforms."
#endif



// CSimpleFunction

class ATL_NO_VTABLE CSimpleFunction :
  public CComObjectRootEx<CComSingleThreadModel>,
  public CComCoClass<CSimpleFunction, &CLSID_SimpleFunction>,
  public IDispatchImpl<ISimpleFunction, &IID_ISimpleFunction, &LIBID_SampleFunctionVS2005Lib, /*wMajor =*/ 1, /*wMinor =*/ 0>,
  public IDispatchImpl<IIadsFunction, &__uuidof(IIadsFunction), &LIBID_IadsFunctionLib, /* wMajor = */ 1, /* wMinor = */ 0>
{
public:
  CSimpleFunction()
  {
  }

  DECLARE_REGISTRY_RESOURCEID(IDR_SIMPLEFUNCTION)


  BEGIN_COM_MAP(CSimpleFunction)
    COM_INTERFACE_ENTRY(ISimpleFunction)
    COM_INTERFACE_ENTRY(IDispatch)
    COM_INTERFACE_ENTRY(IIadsFunction)
  END_COM_MAP()



  DECLARE_PROTECT_FINAL_CONSTRUCT()

  HRESULT FinalConstruct()
  {
    // See SimpleFunction2.h for discussion on this function

    // This variable is for performance and initialization reasons as you will see below in the Compute function. 
    // Make sure to add this member variable to your class -> bool mWasInitialized
    mWasInitialized = false;

    return S_OK;
  }

  void FinalRelease()
  {
    // See SimpleFunction2.h for discussion on this function
  }

  // IIadsFunction Methods
public:

  // See SimpleFunction2.h for a more robust technique in developing a custom function

  STDMETHOD(Compute)( VARIANT * dataIn,  VARIANT * dataOut)
  {
    // About this function
    // FunctionName in Iads -> SampleFunctionVC2005.SimpleFunction  (Basically, this is the "ProgId" for this class)
    // How to call in Iads  -> SampleFunctionVC2005.SimpleFunction( Parameter1, Parameter2, Parameter3 )   (You can also feed in constant values) 
    // In Args              -> FloatValue1, FloatValue2, FloatValue3
    // Out Args             -> FloatValueSum

    // Background info on the Compute function interface
    // The VARIANT "dataIn" argument contains a "SafeArray" of input arguments passed from the user's
    // function (from of their derived equation) into this function. It will contain all the values
    // that you'll need to compute the function result. The result is returned in the "dataOut" argument.
    // The "dataOut" arguement is a VARIANT so we can return numeric results, as well as strings and other types
    // For a more advanced example on returning strings and other types, see the SimpleFunction2 class in this project
    // This function will be called once per data point calculated

    // Get the input arguments array
    VARIANT* argsArray = (VARIANT*)(dataIn->parray->pvData);  // Could use SafeArrayAccessData, but slow..

    // The very first call to this function is designated as an "initializion call". 
    // Check all parameter types here and then never do again. Adding this check to your code will speed up the performance 
    if ( !mWasInitialized )
    {
      // First Step: We need to check how many arguments were passed into our function from the user
      // Example, if the user creates a derived parameter with the function -> SampleFunctionVC2005.SimpleFunction( "Format", ValueOrParam1, ValueOrParam2 )
      // The argument count would be 3
      int argCount = dataIn->parray->rgsabound->cElements;
      if ( argCount != 3 )
      {
        return DISP_E_BADPARAMCOUNT;
      }

      // Second Step: Check Types of each arg..... Either VT_R8 (floating point value), VT_BSTR (string value) for now...
      // In IADS, most every type of numerical argument is passed via an 8 byte floating point value VT_R8. 
      // If you're in doubt, use VT_R8. You can also break here in the code and examine the argsArray[N].vt value to see the actual type 
      // Example, if the user creates a derived parameter with the function -> SampleFunctionVC2005.SimpleFunction( ValueOrParam1, ValueOrParam2, ValueOrParam3 )
      // The argument type for every argument would be VT_R8    
      if ( argsArray[0].vt != VT_R8 ) return E_INVALIDARG;
      if ( argsArray[1].vt != VT_R8 ) return E_INVALIDARG;
      if ( argsArray[2].vt != VT_R8 ) return E_INVALIDARG;

      mWasInitialized = true;
    }

    // Third step: Get the actual values of each arg by extracting from the array of input arguments
    register double p1 = argsArray[0].dblVal;
    register double p2 = argsArray[1].dblVal;
    register double p3 = argsArray[2].dblVal;

    // Final step: Perform your function's purpose and return the output value
    // Add em up and ship it out... Because we're returning a number, the return type is VT_R8 (double) for now... Iads will convert if necessary..
    dataOut->vt = VT_R8;
    dataOut->dblVal = p1 + p2 + p3;

    return S_OK;
  }

private:
  bool mWasInitialized;

};

OBJECT_ENTRY_AUTO(__uuidof(SimpleFunction), CSimpleFunction)
